# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from collections.abc import Sequence
from itertools import count, product
from unittest.mock import Mock

from numpy import array
from pytest import fixture, mark, raises, warns
from qiskit import QuantumCircuit
from qiskit.circuit.random import random_circuit
from qiskit.primitives import EstimatorResult

from zne.extrapolation import Extrapolator, LinearExtrapolator
from zne.noise_amplification import NoiseAmplifier
from zne.noise_amplification.folding_amplifier import MultiQubitAmplifier
from zne.zne_strategy import ZNEStrategy

from . import NO_ITERS_NONE, NO_NONE


################################################################################
## FIXTURES
################################################################################
@fixture(scope="function")
def amplifier_mock():
    """NoiseAmplifier mock object."""
    amplifier = Mock(NoiseAmplifier)
    amplifier.amplify_circuit_noise.side_effect = count()
    return amplifier


@fixture(scope="function")
def extrapolator_mock():
    """Extrapolator mock object."""

    def infer(target, data):
        _, y, _ = zip(*data)
        return 1, 0, {}

    extrapolator = Mock(Extrapolator)
    extrapolator.infer.side_effect = infer
    return extrapolator


################################################################################
## TESTS
################################################################################
def test_definition():
    assert ZNEStrategy._DEFINING_ATTRS == (
        "noise_factors",
        "noise_amplifier",
        "extrapolator",
    )


class TestInit:
    """Test ZNEStrategy initialization logic."""

    def test_defaults(self):
        """Test default configuration."""
        assert ZNEStrategy().noise_amplifier == MultiQubitAmplifier()
        assert ZNEStrategy().noise_factors == (1,)
        assert ZNEStrategy().extrapolator == LinearExtrapolator()

    @mark.parametrize(
        "noise_factors",
        [(1,), (1, 3), (1, 3, 5)],
    )
    def test_custom(
        self,
        noise_factors,
    ):
        """Test custom configuration.

        Proper inputs can be assumed since validation is tested separately.
        """
        noise_amplifier = Mock(NoiseAmplifier)
        extrapolator = Mock(Extrapolator)
        zne_strategy = ZNEStrategy(
            noise_factors=noise_factors,
            noise_amplifier=noise_amplifier,
            extrapolator=extrapolator,
        )
        assert zne_strategy.noise_factors == noise_factors
        assert zne_strategy.noise_amplifier is noise_amplifier
        assert zne_strategy.extrapolator is extrapolator


@mark.parametrize(
    "noise_factors",
    [(1,), (1, 3), (1, 3, 5)],
)
class TestMagic:
    """Test generic ZNEStrategy magic methods."""

    def test_repr(self, noise_factors):
        """Test ZNEStrategy.__repr__() magic method."""
        noise_amplifier = Mock(NoiseAmplifier)
        extrapolator = Mock(Extrapolator)
        zne_strategy = ZNEStrategy(
            noise_factors=noise_factors,
            noise_amplifier=noise_amplifier,
            extrapolator=extrapolator,
        )
        expected = "ZNEStrategy("
        expected += f"noise_factors={repr(noise_factors)}, "
        expected += f"noise_amplifier={repr(noise_amplifier)}, "
        expected += f"extrapolator={repr(extrapolator)})"
        assert repr(zne_strategy) == expected

    def test_eq(self, noise_factors):
        """Test ZNEStrategy.__eq__() magic method."""
        noise_amplifier = Mock(NoiseAmplifier)
        extrapolator = Mock(Extrapolator)
        zne_strategy = ZNEStrategy(
            noise_factors=noise_factors,
            noise_amplifier=noise_amplifier,
            extrapolator=extrapolator,
        )
        assert zne_strategy == ZNEStrategy(
            noise_factors=noise_factors,
            noise_amplifier=noise_amplifier,
            extrapolator=extrapolator,
        )
        assert zne_strategy != ZNEStrategy(
            noise_factors=(*noise_factors, 707),
            noise_amplifier=noise_amplifier,
            extrapolator=extrapolator,
        )
        assert zne_strategy != ZNEStrategy(
            noise_factors=noise_factors,
            noise_amplifier=Mock(NoiseAmplifier),
            extrapolator=extrapolator,
        )
        assert zne_strategy != ZNEStrategy(
            noise_factors=noise_factors,
            noise_amplifier=noise_amplifier,
            extrapolator=Mock(Extrapolator),
        )
        assert zne_strategy != "zne_strategy"

    def test_bool(self, noise_factors):
        noise_amplifier = Mock(NoiseAmplifier)
        extrapolator = Mock(Extrapolator)
        zne_strategy = ZNEStrategy(
            noise_factors=noise_factors,
            noise_amplifier=noise_amplifier,
            extrapolator=extrapolator,
        )
        truth_value = not zne_strategy.is_noop
        assert bool(zne_strategy) is truth_value


class TestConstructors:
    """Test ZNEStrategy constructors."""

    def test_noop(self):
        zne_strategy = ZNEStrategy.noop()
        assert zne_strategy.is_noop


class TestNoiseFactors:
    """Test ZNEStrategy `noise_factors` property."""

    def test_default(self):
        zne_strategy = ZNEStrategy()
        zne_strategy.noise_factors = None
        assert zne_strategy.noise_factors == (1,)

    @mark.parametrize(
        "noise_factors",
        cases := [(1,), (3,), (1, 3), (1, 3, 5), [1, 3, 5], [1.2, 3, 5.4]],
        ids=[f"{nf}" for nf in cases],
    )
    def test_dispatch(self, noise_factors):
        """Test proper noise factors of different types."""
        zne_strategy = ZNEStrategy()
        zne_strategy.noise_factors = noise_factors
        assert zne_strategy.noise_factors == tuple(noise_factors)

    @mark.parametrize(
        "noise_factors, expected",
        cases := list(
            zip(
                [(1, 5, 3), (3.3, 1.2, 5.4)],
                [(1, 3, 5), (1.2, 3.3, 5.4)],
            )
        ),
        ids=[f"{nf}" for nf, _ in cases],
    )
    def test_sort(self, noise_factors, expected):
        """Test unsorted noise factors."""
        zne_strategy = ZNEStrategy()
        with warns(UserWarning):
            zne_strategy.noise_factors = noise_factors
        assert zne_strategy.noise_factors == expected

    @mark.parametrize(
        "noise_factors, expected",
        cases := list(
            zip(
                [(1, 1), (1, 3, 1, 5), (5, 5, 3), (2.4, 2.4)],
                [(1,), (1, 3, 5), (3, 5), (2.4,)],
            )
        ),
        ids=[f"{nf}" for nf, _ in cases],
    )
    def test_duplicates(self, noise_factors, expected):
        """Test duplicate noise factors."""
        zne_strategy = ZNEStrategy()
        with warns(UserWarning):
            zne_strategy.noise_factors = noise_factors
        assert zne_strategy.noise_factors == expected

    @mark.parametrize(
        "noise_factors",
        cases := NO_ITERS_NONE,
        ids=[f"{type(c)}" for c in cases],
    )
    def test_sequence(self, noise_factors):
        """Test type error is raised if noise factors are not Sequence."""
        zne_strategy = ZNEStrategy()
        with raises(TypeError):
            zne_strategy.noise_factors = noise_factors

    @mark.parametrize(
        "noise_factors",
        cases := [(), []],
        ids=[f"{type(c)}" for c in cases],
    )
    def test_empty(self, noise_factors):
        """Test value error is raised for empty lists of noise factors."""
        zne_strategy = ZNEStrategy()
        with raises(ValueError):
            zne_strategy.noise_factors = noise_factors

    @mark.parametrize(
        "noise_factors",
        cases := ["1", True, False, float("NaN"), [1, 3, "5"]],
        ids=[f"{type(c)}" for c in cases],
    )
    def test_real(self, noise_factors):
        """Test type error is raised if noise factors are not real numbers."""
        if not isinstance(noise_factors, Sequence):
            noise_factors = [noise_factors]
        zne_strategy = ZNEStrategy()
        with raises(TypeError):
            zne_strategy.noise_factors = noise_factors

    @mark.parametrize(
        "noise_factors",
        cases := [0, 0.9999, -1, -0.5, (1, 0), (0.9, 1.2)],
        ids=[f"{c}" for c in cases],
    )
    def test_geq_one(self, noise_factors):
        """Test value error is raised if any noise factor is less than one."""
        if not isinstance(noise_factors, Sequence):
            noise_factors = [noise_factors]
        zne_strategy = ZNEStrategy()
        with raises(ValueError):
            zne_strategy.noise_factors = noise_factors


class TestNoiseAmplifier:
    """Test ZNEStrategy `noise_amplifier` property."""

    def test_default(self):
        zne_strategy = ZNEStrategy()
        zne_strategy.noise_amplifier = None
        assert zne_strategy.noise_amplifier == MultiQubitAmplifier()

    @mark.parametrize(
        "noise_amplifier",
        cases := NO_NONE,
        ids=[f"{type(c)}" for c in cases],
    )
    def test_type_error(self, noise_amplifier):
        """Test type error is raised if not `NoiseAmplifier`."""
        zne_strategy = ZNEStrategy()
        with raises(TypeError):
            zne_strategy.noise_amplifier = noise_amplifier


class TestExtrapolator:
    """Test ZNEStrategy `extrapolator` property."""

    def test_default(self):
        zne_strategy = ZNEStrategy()
        zne_strategy.extrapolator = None
        assert zne_strategy.extrapolator == LinearExtrapolator()

    @mark.parametrize(
        "extrapolator",
        cases := NO_NONE,
        ids=[f"{type(c)}" for c in cases],
    )
    def test_type_error(self, extrapolator):
        """Test type error is raised if not `NoiseAmplifier`."""
        zne_strategy = ZNEStrategy()
        with raises(TypeError):
            zne_strategy.extrapolator = extrapolator


class TestProperties:
    """Test generic ZNEStrategy properties."""

    @mark.parametrize("noise_factors", [(1,), (1, 3), (1.2,), (2.1, 4.5)])
    def test_performs_noise_amplification(self, noise_factors):
        """Test if ZNEStrategy performs noise amplification."""
        zne_strategy = ZNEStrategy(noise_factors=noise_factors)
        truth_value = any(nf > 1 for nf in noise_factors)
        if truth_value:
            assert zne_strategy.performs_noise_amplification
        else:
            assert not zne_strategy.performs_noise_amplification

    @mark.parametrize("noise_factors", [(1,), (1, 3), (1.2,), (2.1, 4.5)])
    def test_performs_zne(self, noise_factors):
        """Test if ZNEStrategy performs zero noise extrapolation."""
        zne_strategy = ZNEStrategy(noise_factors=noise_factors)
        truth_value = any(nf > 1 for nf in noise_factors) and len(noise_factors) > 1
        if truth_value:
            assert zne_strategy.performs_zne
        else:
            assert not zne_strategy.performs_zne

    @mark.parametrize("noise_factors", [(1,), (1, 3), (1.2,), (2.1, 4.5)])
    def test_is_noop(self, noise_factors):
        """Test if ZNEStrategy is no-op."""
        zne_strategy = ZNEStrategy(noise_factors=noise_factors)
        truth_value = tuple(noise_factors) == (1,)
        if truth_value:
            assert zne_strategy.is_noop
        else:
            assert not zne_strategy.is_noop


class TestNoiseAmplification:
    """Test ZNEStrategy noise amplification logic."""

    def test_amplify_circuit_noise(self, amplifier_mock):
        noise_factors = (1, 2, 3)
        zne_strategy = ZNEStrategy(noise_factors=noise_factors, noise_amplifier=amplifier_mock)
        circuit = QuantumCircuit(2)
        assert zne_strategy.amplify_circuit_noise(circuit, 1) == 0
        amplifier_mock.amplify_circuit_noise.assert_called_once_with(circuit, 1)
        amplifier_mock.amplify_circuit_noise.reset_mock()
        assert zne_strategy.amplify_circuit_noise(circuit, 1.2) == 1
        amplifier_mock.amplify_circuit_noise.assert_called_once_with(circuit, 1.2)
        circuit.h(0)
        amplifier_mock.amplify_circuit_noise.reset_mock()
        assert zne_strategy.amplify_circuit_noise(circuit, 2.4) == 2
        amplifier_mock.amplify_circuit_noise.assert_called_once_with(circuit, 2.4)

    @mark.parametrize(
        "circuits, noise_factors",
        cases := tuple(
            product(
                [
                    random_circuit(1, 1, seed=0),
                    [],
                    [random_circuit(2, 2, seed=5)],
                    [random_circuit(2, 2, seed=66), random_circuit(2, 2, seed=1081)],
                ],
                [[1], [1, 3], [1, 3, 5]],
            )
        ),
        ids=[f"{type(c).__name__}<{len(c)}>-{nf}" for c, nf in cases],
    )
    def test_build_noisy_circuits(self, amplifier_mock, circuits, noise_factors):
        zne_strategy = ZNEStrategy(noise_factors=noise_factors, noise_amplifier=amplifier_mock)
        _ = zne_strategy.build_noisy_circuits(circuits)
        if isinstance(circuits, QuantumCircuit):
            circuits = [circuits]
        assert amplifier_mock.amplify_circuit_noise.call_count == len(circuits) * len(noise_factors)
        for circuit in circuits:
            for noise_factor in noise_factors:
                amplifier_mock.amplify_circuit_noise.assert_any_call(circuit, noise_factor)

    @mark.parametrize(
        "num_noise_factors, arg, expected",
        cases := [
            (1, None, None),
            (1, 0, (0,)),
            (1, [0], (0,)),
            (1, [1], (1,)),
            (1, [2], (2,)),
            (1, [0, 1], (0, 1)),
            (1, [0, 2], (0, 2)),
            (1, [1, 2], (1, 2)),
            (1, [0, 1, 2], (0, 1, 2)),
            (2, None, None),
            (2, 0, (0, 0)),
            (2, [0], (0, 0)),
            (2, [1], (1, 1)),
            (2, [2], (2, 2)),
            (2, [0, 1], (0, 0, 1, 1)),
            (2, [0, 2], (0, 0, 2, 2)),
            (2, [1, 2], (1, 1, 2, 2)),
            (2, [0, 1, 2], (0, 0, 1, 1, 2, 2)),
            (3, None, None),
            (3, 0, (0, 0, 0)),
            (3, [0], (0, 0, 0)),
            (3, [1], (1, 1, 1)),
            (3, [2], (2, 2, 2)),
            (3, [0, 1], (0, 0, 0, 1, 1, 1)),
            (3, [0, 2], (0, 0, 0, 2, 2, 2)),
            (3, [1, 2], (1, 1, 1, 2, 2, 2)),
            (3, [0, 1, 2], (0, 0, 0, 1, 1, 1, 2, 2, 2)),
        ],
        ids=[f"noise<{nnf}>-{id}" for nnf, id, _ in cases],
    )
    def test_map_to_noisy_circuits(self, num_noise_factors, arg, expected):
        zne_strategy = ZNEStrategy(noise_factors=[n for n in range(1, num_noise_factors + 1)])
        assert zne_strategy.map_to_noisy_circuits(arg) == expected


class TestExtrapolation:
    """Test ZNEStrategy extrapolation logic."""

    @mark.parametrize(
        "noise_factors, values, variances, num_results, extrapolate_return",
        [
            ([1, 2, 3], [1, 2, 3], [0, 0, 0], 1, [0, 1, {}]),
            ([1, 2, 3], [1, 2, 3], [0, 0, 0], 1, [0, 1, {"R2": 0.1}]),
            ([1, 2, 3], [1, 2, 3], [0, 0, 0], 2, [0, 1, {}]),
            ([1, 2, 3], [1, 2, 3], [0, 0, 0], 3, [0, 1, {"R2": 0.1, "P": 6.5}]),
        ],
    )
    def test_mitigate_noisy_result(
        self, noise_factors, values, variances, num_results, extrapolate_return
    ):
        val, err, meta = extrapolate_return
        extrapolator = Mock(Extrapolator)
        extrapolator.extrapolate_zero = Mock(return_value=tuple(extrapolate_return))
        zne_strategy = ZNEStrategy(noise_factors=noise_factors)
        zne_strategy.extrapolator = extrapolator
        metadata = [{"variance": var, "shots": 1024} for var in variances]
        noisy_result = EstimatorResult(
            values=array(values * num_results), metadata=list(metadata * num_results)
        )
        result = zne_strategy.mitigate_noisy_result(noisy_result)
        assert result.values.tolist() == [val] * num_results
        metadatum = {
            "noise_amplification": {
                "noise_amplifier": zne_strategy.noise_amplifier,
                "noise_factors": tuple(noise_factors),
                "values": tuple(values),
                "variance": tuple(variances),
                "shots": tuple([md["shots"] for md in metadata]),
            },
            "extrapolation": {
                "extrapolator": zne_strategy.extrapolator,
                **meta,
            },
        }
        assert result.metadata == [{"std_error": err, "zne": metadatum} for _ in range(num_results)]

    @mark.parametrize(
        "num_noise_factors, values, variances",
        cases := [
            (1, [0], [0]),
            (1, [0, 1], [0, 1]),
            (2, [0, 0], [0, 0]),
            (2, [0, 0, 1, 1], [0, 0, 1, 1]),
            (3, [0, 0, 0], [0, 0, 0]),
            (3, [0, 0, 0, 1, 1, 1], [0, 0, 0, 1, 1, 1]),
        ],
        ids=[f"nf<{nnf}>-val<{len(val)}>" for nnf, val, _ in cases],
    )
    def test_generate_noisy_result_groups(self, num_noise_factors, values, variances):
        zne_strategy = ZNEStrategy(noise_factors=range(1, num_noise_factors + 1))
        metadata = [{"variance": var} for var in variances]
        result = EstimatorResult(values=array(values), metadata=metadata)
        assert (
            len(tuple(zne_strategy._generate_noisy_result_groups(result)))
            == len(values) / num_noise_factors
        )
        for i, group in enumerate(zne_strategy._generate_noisy_result_groups(result)):
            lower = num_noise_factors * i
            upper = num_noise_factors * (i + 1)
            assert group.values.tolist() == values[lower:upper]
            assert group.metadata == metadata[lower:upper]

    @mark.parametrize(
        "num_noise_factors, num_experiments",
        cases := [(2, 1), (2, 3), (3, 2), (3, 4)],
        ids=[f"nf{nnf}-val{ne}" for nnf, ne in cases],
    )
    def test_generate_noisy_result_groups_value_error(self, num_noise_factors, num_experiments):
        zne_strategy = ZNEStrategy(noise_factors=range(1, num_noise_factors + 1))
        values = array([0] * num_experiments)
        metadata = [{"variance": 0.0}] * num_experiments
        result = EstimatorResult(values=values, metadata=metadata)
        with raises(ValueError):
            generator = zne_strategy._generate_noisy_result_groups(result)
            assert next(generator)

    @mark.parametrize(
        "noise_factors, values, std_errors",
        [
            ([1, 2], [0, 1], [0, 0]),
            ([1, 2.0], [0.4, 1.2], [0, None]),
        ],
    )
    def test_regression_data_from_result_group(self, noise_factors, values, std_errors):
        zne_strategy = ZNEStrategy(noise_factors=noise_factors)
        metadatum = {"shots": 1024}
        metadata = [
            {"variance": err**2, **metadatum} if err is not None else metadatum
            for err in std_errors
        ]
        result_group = EstimatorResult(values=array(values), metadata=list(metadata))
        data = zne_strategy._regression_data_from_result_group(result_group)
        expected = (
            noise_factors,
            values,
            [1 for _ in noise_factors],
            [1 if err is None else err for err in std_errors],
        )
        for dat, exp in zip(data, expected):
            assert dat == exp

    @mark.parametrize(
        "num_noise_factors, num_experiments",
        cases := [(1, 2), (2, 1), (3, 1), (2, 3)],
        ids=[f"nf<{nnf}>-experiments<{ne}>" for nnf, ne in cases],
    )
    def test_regression_data_from_result_group_value_error(
        self, num_noise_factors, num_experiments
    ):
        zne_strategy = ZNEStrategy(noise_factors=range(1, num_noise_factors + 1))
        values = [1] * num_experiments
        metadata = [{"variance": 0}] * num_experiments
        result_group = EstimatorResult(values=array(values), metadata=list(metadata))
        with raises(ValueError):
            zne_strategy._regression_data_from_result_group(result_group)

    @mark.parametrize(
        "noise_factors, values, metadata, extrapolation, expected_na",
        [
            (
                [1, 2],
                [1, 1],
                [{"variance": 0}, {"variance": 0}],
                {},
                {"noise_factors": (1, 2), "values": (1, 1), "variance": (0, 0)},
            ),
            (
                [1, 2],
                [1, 0],
                [{"variance": 0.1}, {"variance": 0.4}],
                {"R2": 0.98},
                {"noise_factors": (1, 2), "values": (1, 0), "variance": (0.1, 0.4)},
            ),
            (
                [1, 2, 3],
                [0, 1.5, 2.4],
                [
                    {"variance": 0.11, "shots": 2048},
                    {"variance": 0.1, "shots": 1024},
                    {"variance": 0.12, "shots": 4096},
                ],
                {"R2": 0.44, "P": 6.5},
                {
                    "noise_factors": (1, 2, 3),
                    "values": (0, 1.5, 2.4),
                    "variance": (0.11, 0.1, 0.12),
                    "shots": (2048, 1024, 4096),
                },
            ),
            (
                [1, 2, 3],
                [0, 1.5, 2.4],
                [
                    {"variance": 0.11, "shots": 2048},
                    {"variance": 0.1, "shots": 1024, "backend": "ibmq-nugget"},
                    {"variance": 0.12, "shots": 4096, "seconds": 3600},
                ],
                {"R2": 0.44, "P": 6.5},
                {
                    "noise_factors": (1, 2, 3),
                    "values": (0, 1.5, 2.4),
                    "variance": (0.11, 0.1, 0.12),
                    "shots": (2048, 1024, 4096),
                    "backend": (None, "ibmq-nugget", None),
                    "seconds": (None, None, 3600),
                },
            ),
        ],
    )
    def test_build_zne_metadata(self, noise_factors, values, metadata, extrapolation, expected_na):
        zne_strategy = ZNEStrategy(noise_factors=noise_factors)
        result_group = EstimatorResult(values=array(values), metadata=list(metadata))
        computed = zne_strategy.build_zne_metadata(result_group, extrapolation)
        expected_na = {
            "noise_amplifier": zne_strategy.noise_amplifier,
            **expected_na,
        }
        expected_ex = {
            "extrapolator": zne_strategy.extrapolator,
            **extrapolation,
        }
        assert computed.get("noise_amplification") == expected_na
        assert computed.get("extrapolation") == expected_ex

    @mark.parametrize(
        "num_noise_factors, num_experiments",
        cases := [(1, 2), (2, 1), (3, 1), (2, 3)],
        ids=[f"nf<{nnf}>-experiments<{ne}>" for nnf, ne in cases],
    )
    def test_build_zne_metadata_value_error(self, num_noise_factors, num_experiments):
        zne_strategy = ZNEStrategy(noise_factors=range(1, num_noise_factors + 1))
        values = array([1] * num_experiments)
        metadata = [{"variance": 0}] * num_experiments
        result_group = EstimatorResult(values=values, metadata=metadata)
        with raises(ValueError):
            zne_strategy.build_zne_metadata(result_group)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from test import NO_INTS, NO_ITERS

from pytest import mark, raises

from zne.utils.grouping import (
    from_common_key,
    group_elements,
    merge_dicts,
    squash_for_equal_elements,
)


@mark.parametrize(
    "elements, size, expected",
    cases := [
        ([], 1, []),
        ([], 2, []),
        ([], 3, []),
        ([0], 1, [(0,)]),
        ([0], 2, [(0,)]),
        ([0], 3, [(0,)]),
        ([0, 1], 1, [(0,), (1,)]),
        ([0, 1], 2, [(0, 1)]),
        ([0, 1], 3, [(0, 1)]),
        ([0, 1, 2], 1, [(0,), (1,), (2,)]),
        ([0, 1, 2], 2, [(0, 1), (2,)]),
        ([0, 1, 2], 3, [(0, 1, 2)]),
        ([0, 1, 2, 3], 1, [(0,), (1,), (2,), (3,)]),
        ([0, 1, 2, 3], 2, [(0, 1), (2, 3)]),
        ([0, 1, 2, 3], 3, [(0, 1, 2), (3,)]),
    ],
    ids=[f"list<{len(value)}>-{size}" for value, size, expected in cases],
)
def test_group_elements(elements, size, expected):
    groups = group_elements(elements, group_size=size)
    assert groups == (expected)


@mark.parametrize("elements", NO_ITERS, ids=[str(type(i).__name__) for i in NO_ITERS])
def test_group_elements_type_error_iter(elements):
    with raises(TypeError):
        assert group_elements(elements, group_size=1)


@mark.parametrize("size", NO_INTS, ids=[str(type(i).__name__) for i in NO_INTS])
def test_group_elements_type_error_int(size):
    with raises(TypeError):
        assert group_elements(elements=[], group_size=size)


@mark.parametrize("size", [0, -1])
def test_group_elements_value_error(size):
    with raises(ValueError):
        assert group_elements(elements=[], group_size=size)


@mark.parametrize(
    "dict_list, expected",
    [
        ([{}, {}], {}),
        ([{0: 1}, {}], {0: 1}),
        ([{}, {0: 1}], {0: 1}),
        ([{0: 1}, {0: 1}], {0: 1}),
        ([{0: 2}, {0: 1}], {0: 1}),
        ([{0: 1}, {0: 2}], {0: 2}),
        ([{0: 1}, {2: 3}], {0: 1, 2: 3}),
        ([{2: 3}, {0: 1}], {2: 3, 0: 1}),
        ([{0: 1, 2: 3}, {4: 5}], {0: 1, 2: 3, 4: 5}),
        ([{0: 1, 4: 5}, {2: 3}], {0: 1, 4: 5, 2: 3}),
        ([{4: 5, 0: 1}, {2: 3}], {4: 5, 0: 1, 2: 3}),
        ([{4: 5, 2: 3}, {0: 1}], {4: 5, 2: 3, 0: 1}),
        ([{0: 1, 2: 3}, {0: 2}], {0: 2, 2: 3}),
        ([{4: 5}, {0: 1, 2: 3}], {4: 5, 0: 1, 2: 3}),
        ([{2: 3}, {0: 1, 4: 5}], {2: 3, 0: 1, 4: 5}),
        ([{2: 3}, {4: 5, 0: 1}], {2: 3, 4: 5, 0: 1}),
        ([{0: 1}, {4: 5, 2: 3}], {0: 1, 4: 5, 2: 3}),
        ([{0: 1}, {0: 2, 2: 3}], {0: 2, 2: 3}),
    ],
)
def test_merge_dicts(dict_list, expected):
    assert merge_dicts(dict_list) == expected


@mark.parametrize(
    "dict_list, key, expected",
    tuple(
        zip(
            [
                ({"key": 0}, {"key": 1}),
                ({"key": 0}, {}),
                ({}, {}),
            ],
            ["key", "key", "key"],
            [(0, 1), (0, None), (None, None)],
        )
    ),
)
def test_from_common_key(dict_list, key, expected):
    assert from_common_key(dict_list, key) == expected


@mark.parametrize(
    "sequence, expected",
    tuple(
        zip(
            [(0, 0), (1, 1), (0, 1)],
            [0, 1, (0, 1)],
        )
    ),
)
def test_squash_for_equal_elements(sequence, expected):
    assert squash_for_equal_elements(sequence) == expected


@mark.parametrize("obj", NO_ITERS, ids=[str(type(i).__name__) for i in NO_ITERS])
def test_squash_for_equal_elements_type_error(obj):
    with raises(TypeError):
        assert squash_for_equal_elements(obj)


def test_squash_for_equal_elements_value_error():
    with raises(ValueError):
        assert squash_for_equal_elements([])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Validation utils module."""

from __future__ import annotations

from collections.abc import Callable
from copy import deepcopy
from typing import Any

from .unset import UNSET, UnsetType


class quality:  # pylint: disable=invalid-name
    """Quality attribute similar to property but geared towards custom validation.

    Args:
        fval: function to be used for validating an attribute value on `__set__`
        feff: function to be used for performing side-effects on `__set__` and
            `__delete__` (e.g. erasing cache)
        default: value to assume if set to `...`
        null: value to assume if set to `None`
    """

    __slots__ = "fval", "feff", "_default", "_null", "_name"

    # TODO: update doc to fval's like property does for fget
    def __init__(
        self,
        fval: Callable[[Any, Any], Any] | None = None,
        feff: Callable[[Any], None] | None = None,
        *,
        default: Any = UNSET,
        null: Any = UNSET,
    ) -> None:
        self.fval: Callable[[Any, Any], Any] | None = fval
        self.feff: Callable[[Any], None] | None = feff
        self._default: Any = default
        self._null: Any = null
        self._name: str | None = None

    def __call__(
        self,
        fval: Callable[[Any, Any], Any] | None | UnsetType = UNSET,
        feff: Callable[[Any], None] | None | UnsetType = UNSET,
        *,
        default: Any = UNSET,
        null: Any = UNSET,
    ) -> quality:
        """Returns a shallow copy of the quality with updated attributes."""
        copy = self.__class__(
            fval=fval or self.fval,  # type: ignore
            feff=feff or self.feff,  # type: ignore
            default=default or self.default,
            null=null or self.null,
        )
        copy._name = self.name
        return copy

    def __copy__(self) -> quality:
        return self()

    def __deepcopy__(self, memo: dict) -> quality:
        fval = deepcopy(self.fval, memo)
        feff = deepcopy(self.feff, memo)
        default = deepcopy(self._default, memo)
        null = deepcopy(self._null, memo)
        return self(fval, feff, default=default, null=null)

    ################################################################################
    ## PROPERTIES
    ################################################################################
    @property
    def default(self) -> Any:
        """Default value for the managed attribute to be used for `...`.

        Enforced immutable to avoid side-effects.
        """
        return deepcopy(self._default)

    @property
    def null(self) -> Any:
        """Null value for the managed attribute to be used for `None`.

        Enforced immutable to avoid side-effects.
        """
        return deepcopy(self._null)

    @property
    def name(self) -> str:
        """Name for attr in managed instance as provided by `__set_name__`."""
        return self._name

    @property
    def private_name(self) -> str:
        """Private name to store attr value at in managed instance."""
        if self.name is None:
            return None
        return "__quality_" + self.name

    ################################################################################
    ## DESCRIPTOR PROTOCOL
    ################################################################################
    def __set_name__(self, owner: type, name: str) -> None:
        self._name = name

    def __get__(self, obj: object, objtype: type = None) -> Any:
        if obj is None:
            return self
        if hasattr(obj, self.private_name):
            return getattr(obj, self.private_name)
        setattr(obj, self.name, UNSET)
        return getattr(obj, self.name)

    def __set__(self, obj: object, value: Any) -> None:
        value = self._apply_defaults(value)
        if self.fval is not None:
            value = self.fval(obj, value)
        setattr(obj, self.private_name, value)
        if self.feff is not None:
            self.feff(obj)

    def __delete__(self, obj: object) -> None:
        if hasattr(obj, self.private_name):
            delattr(obj, self.private_name)
        if self.feff is not None:
            self.feff(obj)

    ################################################################################
    ## DECORATORS
    ################################################################################
    def validator(self, fval: Callable[[Any, Any], Any] | None) -> quality:
        """Returns a copy of the quality with a different validator."""
        return self(fval=fval)

    def side_effect(self, feff: Callable[[Any], None] | None) -> quality:
        """Returns a copy of the quality with a different side effect."""
        return self(feff=feff)

    ################################################################################
    ## AUXILIARY
    ################################################################################
    def _apply_defaults(self, value: Any) -> Any:
        """Apply defaults to user input value."""
        default = self.default or self.null
        null = self.null or self.default
        if value is UNSET:
            value = default or None
        elif value is Ellipsis and default is not UNSET:
            value = default
        elif value is None and null is not UNSET:
            value = null
        return value



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from os.path import join as join_path
from tempfile import TemporaryDirectory

from numpy import array
from pytest import mark, raises
from qiskit.primitives import EstimatorResult

from zne.utils.serialization import (
    DumpEncoder,
    EstimatorResultEncoder,
    NumPyEncoder,
    ReprEncoder,
)


@mark.parametrize(
    "obj, expected",
    [
        (0, "0"),
        ([0], "[0]"),
        ((0,), "[0]"),
        ([0, 1], "[0, 1]"),
        ((0, 1), "[0, 1]"),
        (["0", "1"], '["0", "1"]'),
        (("0", "1"), '["0", "1"]'),
        ([[], 1], "[[], 1]"),
        (([], 1), "[[], 1]"),
        ([[0], 1], "[[0], 1]"),
        (([0], 1), "[[0], 1]"),
        ({"key": "value"}, '{"key": "value"}'),
        ({0: 1}, '{"0": 1}'),
        ({0: (1,)}, '{"0": [1]}'),
        ({0: [1]}, '{"0": [1]}'),
    ],
)
class TestDumpEncoder:
    def test_dump(self, obj, expected):
        with TemporaryDirectory() as tmpdir:
            file_path = join_path(tmpdir, "zne-dump")
            DumpEncoder.dump(obj, file=file_path)
            with open(file_path) as f:
                contents = f.read()
            assert contents == expected

    def test_dumps(self, obj, expected):
        assert DumpEncoder.dumps(obj) == expected


class TestReprEncoder:
    @mark.parametrize(
        "repr_str",
        cases := [
            "",
            "dummy",
            "some spaces here",
            "`~!@#$%^&*()-_=+[]\\\"{}|;:',<.>/?",
        ],
        ids=cases,
    )
    def test_default(self, repr_str):
        class DummyRepr:
            def __repr__(self):
                return repr_str

        obj = DummyRepr()
        enc = ReprEncoder()
        assert enc.default(obj) == repr(obj)


class TestNumPyEncoder:
    @mark.parametrize(
        "array_like",
        cases := [
            [0, 1, 2],
            (0, 1, 2),
        ],
        ids=[type(c) for c in cases],
    )
    def test_default(self, array_like):
        a = array(array_like)
        enc = NumPyEncoder()
        assert enc.default(a) == a.tolist()

    def test_default_type_error(self):
        with raises(TypeError):
            _ = NumPyEncoder().default({"call": "super"})


class TestEstimatorResultEncoder:
    @mark.parametrize(
        "values, metadata",
        zip(
            [
                array([]),
                array([1]),
                array([1, 2]),
            ],
            [
                [],
                [{"variance": 0}],
                [{"variance": 0}, {"variance": 1}],
            ],
        ),
    )
    def test_default(self, values, metadata):
        result = EstimatorResult(values=values, metadata=metadata)
        enc = EstimatorResultEncoder()
        assert enc.default(result) == {"values": result.values, "metadata": result.metadata}

    def test_numpy_subclass(self):
        enc = EstimatorResultEncoder()
        assert isinstance(enc, NumPyEncoder)
        a = array([0, 1, 2])
        assert enc.default(a) == a.tolist()

    def test_repr_subclass(self):
        enc = EstimatorResultEncoder()
        assert isinstance(enc, ReprEncoder)
        obj = {"call": "super"}
        assert enc.default(obj) == repr(obj)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Type definitions for the Zero Noise Extrapolation (ZNE) Estimator class."""

from typing import Any, Dict, Sequence, Tuple, Union

from qiskit import QuantumCircuit

Metadata = Dict[str, Any]
EstimatorResultData = Dict[str, Union[float, Metadata]]
ParameterVector = Sequence[float]  # TODO: qiskit.circuit::ParameterVector
CircuitKey = tuple
NoiseFactor = float
ZNECacheKey = Tuple[CircuitKey, NoiseFactor]
ZNECache = Dict[ZNECacheKey, QuantumCircuit]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from test import NO_INTS, NO_REAL, TYPES
from unittest.mock import Mock, patch

from numpy.random import Generator, default_rng
from pytest import fixture, mark, raises, warns
from qiskit import QuantumCircuit

from zne.noise_amplification.folding_amplifier.global_folding_amplifier import (
    GlobalFoldingAmplifier,
)
from zne.noise_amplification.folding_amplifier.local_folding_amplifier import (
    LocalFoldingAmplifier,
)

MOCK_TARGET_PATH = "zne.noise_amplification.folding_amplifier.folding_amplifier.FoldingAmplifier"


@fixture(scope="module")
def patch_amplifier_with_multiple_mocks():
    def factory_method(**kwargs):
        return patch.multiple(MOCK_TARGET_PATH, **kwargs)

    return factory_method


@mark.parametrize(
    "NoiseAmplifier",
    (GlobalFoldingAmplifier, LocalFoldingAmplifier),
    ids=("GlobalFoldingAmplifier", "LocalFoldingAmplifier"),
)
class TestFoldingAmplifier:
    @fixture(scope="function")
    def setter_mocks(self):
        mocks = {
            "_set_sub_folding_option": Mock(),
            "_set_barriers": Mock(),
            "_prepare_rng": Mock(),
            "_set_noise_factor_relative_tolerance": Mock(),
        }
        return mocks

    ################################################################################
    ## INIT TESTS
    ################################################################################

    def test_init_default_kwargs(
        self, NoiseAmplifier, patch_amplifier_with_multiple_mocks, setter_mocks
    ):
        with patch_amplifier_with_multiple_mocks(**setter_mocks):
            NoiseAmplifier()
        setter_mocks["_set_sub_folding_option"].assert_called_once_with("from_first")
        setter_mocks["_set_barriers"].assert_called_once_with(True)
        setter_mocks["_prepare_rng"].assert_called_once_with(None)
        setter_mocks["_set_noise_factor_relative_tolerance"].assert_called_once_with(1e-2)

    def test_init_custom_kwargs(
        self, NoiseAmplifier, patch_amplifier_with_multiple_mocks, setter_mocks
    ):
        with patch_amplifier_with_multiple_mocks(**setter_mocks):
            NoiseAmplifier(
                sub_folding_option="random",
                barriers=False,
                random_seed=1,
                noise_factor_relative_tolerance=1e-1,
            )
        setter_mocks["_set_sub_folding_option"].assert_called_once_with("random")
        setter_mocks["_set_barriers"].assert_called_once_with(False)
        setter_mocks["_prepare_rng"].assert_called_once_with(1)
        setter_mocks["_set_noise_factor_relative_tolerance"].assert_called_once_with(1e-1)

    ################################################################################
    ## PROPERTIES and SETTER TESTS
    ################################################################################

    @mark.parametrize(
        "sub_folding_option",
        cases := ["from_last", "from_first", "random"],
        ids=[f"{c}" for c in cases],
    )
    def test_set_sub_folding_option(self, NoiseAmplifier, sub_folding_option):
        noise_amplifier = NoiseAmplifier()
        noise_amplifier._set_sub_folding_option(sub_folding_option)
        assert noise_amplifier.sub_folding_option == sub_folding_option

    @mark.parametrize(
        "sub_folding_option", [(t,) for t in TYPES], ids=[str(type(i).__name__) for i in TYPES]
    )
    def test_set_sub_folding_option_value_error(self, NoiseAmplifier, sub_folding_option):
        with raises(ValueError):
            NoiseAmplifier()._set_sub_folding_option(sub_folding_option)

    @mark.parametrize(
        "random_seed",
        cases := [1, 2],
        ids=[f"{c}" for c in cases],
    )
    def test_prepare_rng(self, NoiseAmplifier, random_seed):
        noise_amplifier = NoiseAmplifier()
        noise_amplifier._prepare_rng(random_seed)
        assert isinstance(noise_amplifier._rng, Generator)
        rng = default_rng(random_seed)
        assert noise_amplifier._rng.bit_generator.state == rng.bit_generator.state

    @mark.parametrize(
        "random_seed", [(t,) for t in NO_INTS], ids=[str(type(i).__name__) for i in NO_INTS]
    )
    def test_prepare_rng_type_error(self, NoiseAmplifier, random_seed):
        with raises(TypeError):
            NoiseAmplifier()._prepare_rng(random_seed)

    @mark.parametrize(
        "tolerance",
        cases := [0.1, 0.01],
        ids=[f"{c}" for c in cases],
    )
    def test_set_noise_factor_relative_tolerance(self, NoiseAmplifier, tolerance):
        noise_amplifier = NoiseAmplifier()
        noise_amplifier._set_noise_factor_relative_tolerance(tolerance)
        assert noise_amplifier._noise_factor_relative_tolerance == tolerance

    @mark.parametrize(
        "tolerance", [(t,) for t in NO_REAL], ids=[str(type(i).__name__) for i in NO_REAL]
    )
    def test_set_noise_factor_relative_tolerance_type_error(self, NoiseAmplifier, tolerance):
        with raises(TypeError):
            NoiseAmplifier()._set_noise_factor_relative_tolerance(tolerance)

    @mark.parametrize("warn_user", cases := [True, False], ids=[f"{c}" for c in cases])
    def test_set_warn_user(self, NoiseAmplifier, warn_user):
        noise_amplifier = NoiseAmplifier()
        noise_amplifier.warn_user = warn_user
        assert noise_amplifier.warn_user == warn_user

    @mark.parametrize(
        "warn_user",
        cases := [t for t in TYPES if not isinstance(t, bool)],
        ids=[str(type(c).__name__) for c in cases],
    )
    def test_set_warn_user_type_error(self, NoiseAmplifier, warn_user):
        noise_amplifier = NoiseAmplifier()
        with raises(TypeError):
            noise_amplifier.warn_user = warn_user

    @mark.parametrize(
        "barriers", cases := [True, False, 1, 0, "", "|"], ids=[f"{c}" for c in cases]
    )
    def test_set_barriers(self, NoiseAmplifier, barriers):
        noise_amplifier = NoiseAmplifier()
        noise_amplifier._set_barriers(barriers)
        assert noise_amplifier.barriers is bool(barriers)

    ################################################################################
    ## TESTS
    ################################################################################

    @mark.parametrize(
        "noise_factor",
        cases := [1, 1.2, 2, 3.5],
        ids=[f"{c}" for c in cases],
    )
    def test_validate_noise_factor(self, NoiseAmplifier, noise_factor):
        NoiseAmplifier()._validate_noise_factor(noise_factor)

    @mark.parametrize(
        "noise_factor",
        cases := [0, 0.5, -1, -2.5],
        ids=[f"{c}" for c in cases],
    )
    def test_validate_noise_factor_value_error(self, NoiseAmplifier, noise_factor):
        with raises(ValueError):
            NoiseAmplifier()._validate_noise_factor(noise_factor)

    @mark.parametrize(
        "folding, expected",
        cases := tuple(
            zip(
                [0, 1, 2],
                [1, 3, 5],
            )
        ),
        ids=[f"{f}-{e}" for f, e in cases],
    )
    def test_folding_to_noise_factor(self, NoiseAmplifier, folding, expected):
        assert NoiseAmplifier().folding_to_noise_factor(folding) == expected

    def test_warn_user_true(self, NoiseAmplifier):
        with warns(UserWarning):
            NoiseAmplifier().warn("warning")

    def test_warn_user_false(self, NoiseAmplifier):
        NoiseAmplifier(warn_user=False).warn("no warning")

    @mark.parametrize(
        "num_instructions, noise_factor, expected",
        cases := [
            (9, 1.2, (0, 1)),
            (9, 2, (0, 4)),
            (9, 3.6, (1, 3)),
            (9, 4, (1, 5)),
            (9, 17 / 3.0, (2, 3)),
            (9, 100, (49, 5)),
            (17, 1.2, (0, 2)),
            (17, 2, (0, 8)),
            (17, 3.6, (1, 5)),
            (17, 4, (1, 9)),
            (17, 100, (49, 9)),
        ],
        ids=[f"{ni}-{nf}-{cd}" for ni, nf, cd in cases],
    )
    @mark.filterwarnings("ignore::UserWarning")
    def test_compute_folding_nums(self, NoiseAmplifier, num_instructions, noise_factor, expected):
        assert NoiseAmplifier()._compute_folding_nums(noise_factor, num_instructions) == expected

    @mark.parametrize(
        "noise_factor",
        cases := [1.5, 3.05],
        ids=[f"{c}" for c in cases],
    )
    def test_compute_folding_nums_warning(self, NoiseAmplifier, noise_factor):
        with warns(UserWarning):
            assert NoiseAmplifier()._compute_folding_nums(noise_factor, 3)

    def test_compute_folding_no_foldings(self, NoiseAmplifier):
        with warns(UserWarning):
            assert NoiseAmplifier()._compute_folding_nums(3, 0) == (0, 0)

    @mark.parametrize("registers", [(), (0,), (1,), (0, 1)])
    def test_apply_barrier_true(self, NoiseAmplifier, registers):
        circuit = QuantumCircuit(2)
        original = circuit.copy()
        noise_amplifier = NoiseAmplifier(barriers=True)
        circuit = noise_amplifier._apply_barrier(circuit, *registers)
        last_instruction = circuit.data.pop()
        assert last_instruction.operation.name == "barrier"
        registers = [circuit.qubits[r] for r in registers] or circuit.qubits
        assert last_instruction.qubits == tuple(registers)
        assert circuit == original

    @mark.parametrize("registers", [(), (0,), (1,), (0, 1)])
    def test_apply_barrier_false(self, NoiseAmplifier, registers):
        circuit = QuantumCircuit(2)
        original = circuit.copy()
        noise_amplifier = NoiseAmplifier(barriers=False)
        circuit = noise_amplifier._apply_barrier(circuit, *registers)
        assert circuit == original



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Noise amplification strategy via unitary/inverse repetition."""

from .global_folding_amplifier import GlobalFoldingAmplifier
from .local_folding_amplifier import (
    CxAmplifier,
    LocalFoldingAmplifier,
    MultiQubitAmplifier,
    TwoQubitAmplifier,
)

__all__ = [
    "GlobalFoldingAmplifier",
    "LocalFoldingAmplifier",
    "CxAmplifier",
    "TwoQubitAmplifier",
    "MultiQubitAmplifier",
]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from itertools import product
from test import ITERS, NO_INTS_NONE
from unittest.mock import Mock, call, patch

from numpy.random import default_rng
from pytest import approx, fixture, mark, raises, warns
from qiskit.circuit import CircuitInstruction, ParameterVector, QuantumCircuit
from qiskit.circuit.barrier import Barrier
from qiskit.circuit.library.standard_gates import CCXGate, CXGate, HGate, RZGate, XGate
from qiskit.circuit.random import random_circuit
from qiskit.quantum_info.operators import Operator
from qiskit.transpiler.passes import RemoveBarriers

from zne.noise_amplification import (
    CxAmplifier,
    LocalFoldingAmplifier,
    MultiQubitAmplifier,
    TwoQubitAmplifier,
)

MOCK_TARGET_PATH = (
    "zne.noise_amplification.folding_amplifier.local_folding_amplifier.LocalFoldingAmplifier"
)


################################################################################
## MODULE FIXTURES
################################################################################
@fixture(scope="module")
def noise_amplifier():
    return LocalFoldingAmplifier()


@fixture(scope="module")
def noise_factor():
    return 3


@fixture(scope="module")
def patch_amplifier_with_mock():
    def factory_method(method_name=None, **kwargs):
        method_name = method_name or ""
        return patch(MOCK_TARGET_PATH + method_name, **kwargs)

    return factory_method


@fixture(scope="module")
def patch_amplifier_with_multiple_mocks(**kwargs):
    def factory_method(**kwargs):
        return patch.multiple(MOCK_TARGET_PATH, **kwargs)

    return factory_method


################################################################################
## INIT TESTS
################################################################################
def test_init_default_kwargs(patch_amplifier_with_mock):
    with patch_amplifier_with_mock(method_name="._set_gates_to_fold") as mock:
        LocalFoldingAmplifier()
    mock.assert_called_once_with(None)


def test_init_custom_kwargs(patch_amplifier_with_mock):
    with patch_amplifier_with_mock(method_name="._set_gates_to_fold") as mock:
        LocalFoldingAmplifier(gates_to_fold=2)
    mock.assert_called_once_with(2)


################################################################################
## PROPERTIES and SETTER TESTS
################################################################################
@mark.parametrize("gates_to_fold", cases := [1, 2, 3], ids=[f"{c}" for c in cases])
def test_set_gates_to_fold_int(gates_to_fold):
    noise_amplifier = LocalFoldingAmplifier()
    noise_amplifier._set_gates_to_fold(gates_to_fold)
    assert noise_amplifier.gates_to_fold == frozenset((gates_to_fold,))


@mark.parametrize("gates_to_fold", cases := ["h", "cx", "ccx"], ids=[f"{c}" for c in cases])
def test_set_gates_to_fold_str(gates_to_fold):
    noise_amplifier = LocalFoldingAmplifier()
    noise_amplifier._set_gates_to_fold(gates_to_fold)
    assert noise_amplifier.gates_to_fold == frozenset((gates_to_fold,))


@mark.parametrize(
    "gates_to_fold",
    cases := [["x"], (2,), ["h", 2, "ccx"], ("rz", 1, 3, "i")],
    ids=[f"{c}" for c in cases],
)
def test_set_gates_to_fold_sequence(gates_to_fold):
    noise_amplifier = LocalFoldingAmplifier()
    noise_amplifier._set_gates_to_fold(gates_to_fold)
    assert noise_amplifier.gates_to_fold == frozenset(gates_to_fold)


def test_set_gates_to_fold_none(noise_amplifier):
    noise_amplifier = LocalFoldingAmplifier()
    noise_amplifier._set_gates_to_fold(None)
    assert noise_amplifier.gates_to_fold is None


@mark.parametrize(
    "gates_to_fold",
    cases := [t for t in NO_INTS_NONE if t not in ITERS],
    ids=[str(type(t).__name__) for t in cases],
)
def test_set_gates_to_fold_type_error(gates_to_fold):
    with raises(TypeError):
        LocalFoldingAmplifier()._set_gates_to_fold(gates_to_fold)


@mark.parametrize(
    "gates_to_fold",
    cases := [(2, 1.0), ("h", 2, None), ("rz", 1j, 3, "i")],
    ids=[f"{c}" for c in cases],
)
def test_validate_gates_to_fold_type_error(gates_to_fold):
    with raises(TypeError):
        LocalFoldingAmplifier()._validate_gates_to_fold(frozenset(gates_to_fold))


@mark.parametrize("gates_to_fold", cases := [(1, 0), (-2, "h")], ids=[f"{c}" for c in cases])
def test_validate_gates_to_fold_value_error(gates_to_fold):
    with raises(ValueError):
        LocalFoldingAmplifier()._validate_gates_to_fold(frozenset(gates_to_fold))


@mark.parametrize(
    "gates_to_fold", cases := [(1, "areeq"), ("pedro", "h")], ids=[f"{c}" for c in cases]
)
def test_validate_gates_to_fold_warning(gates_to_fold):
    with warns(UserWarning):
        LocalFoldingAmplifier()._validate_gates_to_fold(frozenset(gates_to_fold))


################################################################################
## PUBLIC METHODS TESTS
################################################################################
def test_amplify_circuit_noise(
    patch_amplifier_with_multiple_mocks, noise_amplifier, get_random_circuit, circuit, noise_factor
):
    # TODO: new test after refactor
    # TODO: test barriers are added
    example_foldings = list(range(len(circuit)))
    example_circuits = [get_random_circuit(i) for i in range(len(circuit) + 1)]
    mocks = {
        "_validate_noise_factor": Mock(),
        "_build_foldings_per_gate": Mock(return_value=example_foldings),
        "_append_folded": Mock(side_effect=example_circuits[1:]),
    }
    circuit_mock_target = (
        "zne.noise_amplification.folding_amplifier.local_folding_amplifier"
        ".QuantumCircuit.copy_empty_like"
    )
    with patch_amplifier_with_multiple_mocks(**mocks):
        with patch(circuit_mock_target, return_value=example_circuits[0]) as _:
            _ = noise_amplifier.amplify_circuit_noise(circuit, noise_factor)
    mocks["_validate_noise_factor"].assert_called_once_with(noise_factor)
    mocks["_build_foldings_per_gate"].assert_called_once_with(circuit, noise_factor)
    # circuit_copy_mock.assert_called_once_with()
    calls = [
        call(c, o, f)
        for c, o, f in zip(example_circuits[:-1], circuit.data, example_foldings)
        if f != 0
    ]
    assert mocks["_append_folded"].call_count == len(calls)
    # mocks["_append_folded"].assert_has_calls(calls, any_order=False)
    # assert noisy_circuit == example_circuits[-1]


################################################################################
## PRIVATE METHODS TESTS
################################################################################
@mark.parametrize(
    "foldings, mask, expected",
    cases := [
        ([], [False, False, False], [0, 0, 0]),
        ([5, 6], [True, True], [5, 6]),
        ([2, 1, 2], [False, False, True, False, True, True], [0, 0, 2, 0, 1, 2]),
    ],
    ids=[f"{len(f)}-{len(m)}-{len(e)}" for f, m, e in cases],
)
def test_build_foldings_per_gate(
    noise_amplifier,
    patch_amplifier_with_multiple_mocks,
    foldings,
    mask,
    expected,
):
    noise_factor = 3
    circuit = QuantumCircuit(2)
    for i, _ in enumerate(mask):
        circuit.h(i % circuit.num_qubits)
    mocks = {
        "_check_gate_folds": Mock(side_effect=mask),
        "_build_foldings": Mock(return_value=foldings),
    }
    with patch_amplifier_with_multiple_mocks(**mocks):
        gate_foldings = noise_amplifier._build_foldings_per_gate(circuit, noise_factor)
    calls = [call(operation) for operation in circuit.data]
    assert mocks["_check_gate_folds"].call_count == len(calls)
    mocks["_check_gate_folds"].assert_has_calls(calls, any_order=False)
    mocks["_build_foldings"].assert_called_once_with(noise_factor, mask.count(True))
    assert len(gate_foldings) == len(mask)
    assert gate_foldings == expected


@mark.parametrize(
    "folding_nums, sub_foldings",
    cases := [
        ([0, 0], []),
        ([2, 3], [0, 0, 1, 0, 1, 1]),
    ],
    ids=[f"{len(f)}-{len(m)}" for f, m in cases],
)
def test_build_foldings(
    noise_amplifier,
    patch_amplifier_with_multiple_mocks,
    folding_nums,
    sub_foldings,
):
    num_full_foldings, num_sub_foldings = folding_nums
    num_gates_to_fold = len(sub_foldings)
    noise_factor = 3
    mocks = {
        "_compute_folding_nums": Mock(return_value=folding_nums),
        "_build_sub_foldings": Mock(return_value=sub_foldings),
    }
    with patch_amplifier_with_multiple_mocks(**mocks):
        foldings = noise_amplifier._build_foldings(noise_factor, num_gates_to_fold)
    mocks["_compute_folding_nums"].assert_called_once_with(noise_factor, num_gates_to_fold)
    mocks["_build_sub_foldings"].assert_called_once_with(num_sub_foldings, num_gates_to_fold)
    assert foldings == [num_full_foldings + sf for sf in sub_foldings]


class TestCheckGateFolds:
    @mark.parametrize("num_qubits", cases := [0, 1, 3], ids=[f"{c}" for c in cases])
    def test_check_gate_folds_barrier(self, noise_amplifier, num_qubits):
        operation = CircuitInstruction(Barrier(num_qubits), tuple(range(num_qubits)), ())
        assert noise_amplifier._check_gate_folds(operation) is False

    @mark.parametrize(
        "operation",
        cases := [
            CircuitInstruction(HGate(), (1,), ()),
            CircuitInstruction(CXGate(), (0, 1), ()),
            CircuitInstruction(RZGate(0), (0, 1), ()),
            CircuitInstruction(XGate(), (0,), ()),
            CircuitInstruction(CCXGate(), (0, 1, 2), ()),
        ],
        ids=[f"<{o[0].name}>" for o in cases],
    )
    def test_check_gate_folds_none(self, noise_amplifier, operation):
        noise_amplifier = LocalFoldingAmplifier(gates_to_fold=None)
        assert noise_amplifier._check_gate_folds(operation) is True

    @mark.parametrize(
        "operation, gates_to_fold, expected",
        cases := [
            (CircuitInstruction(HGate(), (1,), ()), 1, True),
            (CircuitInstruction(CXGate(), (0, 1), ()), [1, 2], True),
            (CircuitInstruction(RZGate(0), (0, 1), ()), 1, False),
            (CircuitInstruction(XGate(), (0,), ()), 2, False),
            (CircuitInstruction(CCXGate(), (0, 1, 2), ()), [2, 3], True),
            (CircuitInstruction(CCXGate(), (0, 1, 2), ()), [1, 2], False),
        ],
        ids=[f"<{o[0].name}>-{gtf}-{e}" for o, gtf, e in cases],
    )
    def test_check_gate_folds_num_qubits(self, operation, gates_to_fold, expected):
        noise_amplifier = LocalFoldingAmplifier(gates_to_fold=gates_to_fold)
        assert noise_amplifier._check_gate_folds(operation) is expected

    @mark.parametrize(
        "operation, gates_to_fold, expected",
        cases := [
            (CircuitInstruction(HGate(), (1,), ()), "h", True),
            (CircuitInstruction(CXGate(), (0, 1), ()), ["h", "cx"], True),
            (CircuitInstruction(RZGate(0), (0, 1), ()), "h", False),
            (CircuitInstruction(XGate(), (0,), ()), ["h", "cx"], False),
            (CircuitInstruction(CCXGate(), (0, 1, 2), ()), "ccx", True),
        ],
        ids=[f"<{o[0].name}>-{gtf}-{e}" for o, gtf, e in cases],
    )
    def test_check_gate_folds_instruction_names(self, operation, gates_to_fold, expected):
        noise_amplifier = LocalFoldingAmplifier(gates_to_fold=gates_to_fold)
        assert noise_amplifier._check_gate_folds(operation) is expected


@mark.parametrize(
    "num_foldings, num_instructions",
    cases := tuple(
        product(
            [0, 1, 3],
            [3, 7],
        )
    ),
    ids=[f"{nf}-{ni}" for nf, ni in cases],
)
class TestBuildSubFoldings:
    def test_sub_foldings_from_first(self, num_foldings, num_instructions):
        noise_amplifier = LocalFoldingAmplifier(sub_folding_option="from_first")
        sub_foldings_list = noise_amplifier._build_sub_foldings(num_foldings, num_instructions)
        assert len(sub_foldings_list) == num_instructions
        assert sum(sub_foldings_list) == num_foldings
        assert sub_foldings_list == [1] * num_foldings + [0] * (num_instructions - num_foldings)

    def test_sub_foldings_from_last(self, num_foldings, num_instructions):
        noise_amplifier = LocalFoldingAmplifier(sub_folding_option="from_last")
        sub_foldings_list = noise_amplifier._build_sub_foldings(num_foldings, num_instructions)
        assert len(sub_foldings_list) == num_instructions
        assert sum(sub_foldings_list) == num_foldings
        assert sub_foldings_list == [0] * (num_instructions - num_foldings) + [1] * num_foldings

    @mark.parametrize("seed", cases := [1, 5, 66], ids=[f"{c}" for c in cases])
    def test_build_sub_foldings_random(self, num_foldings, num_instructions, seed):
        noise_amplifier = LocalFoldingAmplifier(sub_folding_option="random", random_seed=seed)
        sub_foldings_list = noise_amplifier._build_sub_foldings(num_foldings, num_instructions)
        assert len(sub_foldings_list) == num_instructions
        assert sum(sub_foldings_list) == num_foldings
        idxs = default_rng(seed).choice(num_instructions, size=num_foldings, replace=False)
        assert sub_foldings_list == [1 if i in idxs else 0 for i in range(num_instructions)]


@mark.parametrize(
    "seed , num_foldings",
    cases := tuple(
        product(
            [1, 2, 3],
            [0, 1, 3],
        )
    ),
    ids=[f"Instruction<{s}>-{nf}" for s, nf in cases],
)
def test_append_folded(noise_amplifier, circuit, get_random_circuit, seed, num_foldings):
    operation = get_random_circuit(seed).data[0]
    noisy_circuit = noise_amplifier._append_folded(circuit.copy(), operation, num_foldings)
    assert noisy_circuit.data[: len(circuit)] == circuit.data
    operation_gen = (operation for operation in noisy_circuit[len(circuit) :])
    instruction, qargs, cargs = operation
    barrier = CircuitInstruction(Barrier(len(qargs)), qargs, [])
    inverse = CircuitInstruction(instruction.inverse(), qargs, cargs)
    for _ in range(num_foldings):
        assert next(operation_gen) == barrier
        assert next(operation_gen) == operation
        assert next(operation_gen) == barrier
        assert next(operation_gen) == inverse
    assert next(operation_gen) == barrier
    assert next(operation_gen) == operation
    assert next(operation_gen) == barrier


@mark.parametrize("num_foldings", [0.0, 0.5, 1.0, -1.0])
def test_append_folded_type_error(circuit, noise_amplifier, num_foldings):
    with raises(TypeError):
        noise_amplifier._append_folded(circuit, circuit.data[0], num_foldings)


@mark.parametrize("num_foldings", [-1, -2])
def test_append_folded_value_error(circuit, noise_amplifier, num_foldings):
    with raises(ValueError):
        noise_amplifier._append_folded(circuit, circuit.data[0], num_foldings)


class TestParametrizedCircuits:
    @fixture(scope="class")
    def parametrized_circuit(self):
        circuit = QuantumCircuit(2)
        params = ParameterVector("P", 4)
        circuit.ry(params[0], 0)
        circuit.rz(params[0], 1)
        circuit.crx(params[1], 0, 1)
        circuit.rz(params[2], 0)
        circuit.cu(params[3], params[0], params[1], params[0], 1, 0)
        circuit.ry(params[3], 1)
        circuit.crx(params[3], 0, 1)
        return circuit

    def test_append_folded_parameters(self, noise_amplifier, parametrized_circuit, noise_factor):
        example_operation = parametrized_circuit.data[-1]
        noisy_circuit = noise_amplifier._append_folded(
            parametrized_circuit.copy(), example_operation, noise_factor
        )
        instruction_gen = (
            instruction for instruction, _, _ in noisy_circuit[len(parametrized_circuit) :]
        )
        original_instruction, _, _ = example_operation
        for i in range(noise_factor):
            if i % 2 == 0:
                next(instruction_gen)
                assert next(instruction_gen).params == original_instruction.params
            else:
                next(instruction_gen)
                assert next(instruction_gen).params == original_instruction.inverse().params


################################################################################
## INTEGRATION TESTS
################################################################################
@mark.parametrize(
    "circuit, noise_factor, gates_to_fold, sub_folding_option",
    cases := tuple(
        product(
            [
                random_circuit(5, 3, seed=0),
                random_circuit(5, 3, seed=5),
                random_circuit(5, 6, seed=66),
                random_circuit(5, 6, seed=1081),
            ],
            [1.5, 3.3, 17 / 3],
            [None, 1, [2, "h", "rz", "u2", "u3"]],
            ["from_last", "from_first", "random"],
        )
    ),
    ids=[f"{type(c).__name__}<{len(c)}>-{nf}-{gtf}-{sfo}" for c, nf, gtf, sfo in cases],
)
class TestIntegration:
    @mark.filterwarnings("ignore::UserWarning")
    def test_noise_factor_of_folded_circuit(
        self, circuit, noise_factor, gates_to_fold, sub_folding_option
    ):
        noise_amplifier = LocalFoldingAmplifier(
            gates_to_fold=gates_to_fold, sub_folding_option=sub_folding_option
        )
        noisy_circuit = noise_amplifier.amplify_circuit_noise(circuit.copy(), noise_factor)
        noisy_circuit = RemoveBarriers()(noisy_circuit)
        num_gates_to_fold = [
            noise_amplifier._check_gate_folds(operation) for operation in circuit.data
        ].count(True)
        num_instructions_folded = [
            noise_amplifier._check_gate_folds(operation) for operation in noisy_circuit.data
        ].count(True)
        num_foldings = round(num_gates_to_fold * (noise_factor - 1) / 2.0)
        closest_noise_factor = 2 * num_foldings / num_gates_to_fold + 1
        assert num_instructions_folded / num_gates_to_fold == approx(closest_noise_factor)
        assert Operator(noisy_circuit).equiv(Operator(circuit))


################################################################################
## FACADES
################################################################################
class TestFacades:
    @mark.parametrize(
        "cls, configs",
        [
            (CxAmplifier, {"gates_to_fold": "cx"}),
            (TwoQubitAmplifier, {"gates_to_fold": 2}),
        ],
    )
    def test_amplifier_facade(self, cls, configs):
        amplifier = cls()
        assert isinstance(amplifier, LocalFoldingAmplifier)
        for key, value in configs.items():
            assert getattr(amplifier, key) == frozenset([value])

    def test_multi_qubit_amplifier(self):
        amplifier = MultiQubitAmplifier()
        assert isinstance(amplifier, LocalFoldingAmplifier)
        assert amplifier.gates_to_fold is None
        circuit = QuantumCircuit(4)
        circuit.h(0)
        circuit.z(3)
        circuit.cx(0, 1)
        circuit.x(1)
        circuit.ccx(0, 1, 2)
        circuit.y(2)
        circuit.ecr(2, 3)
        circuit.measure_all()
        for operation in circuit.data:
            instruction = operation[0]
            skipped_instructions = {"barrier", "measure"}
            expected = instruction.num_qubits > 1 and instruction.name not in skipped_instructions
            assert amplifier._check_gate_folds(operation) == expected



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from unittest.mock import Mock

from pytest import fixture, mark, raises

from zne.meta.init import zne_init
from zne.zne_strategy import ZNEStrategy


class TestZNEInit:
    ################################################################################
    ## FIXTURES
    ################################################################################
    @fixture(scope="function")
    def init(_self):
        return zne_init(Mock())

    @fixture(scope="function")
    def self(_self):
        return Mock()

    ################################################################################
    ## TESTS
    ################################################################################
    def test_deprecation(_self, init):
        with raises(TypeError):
            init("self", circuits="circuits")
        with raises(TypeError):
            init("self", observables="observables")
        with raises(TypeError):
            init("self", parameters="parameters")

    def test_base_init(_self, self):
        base_init = Mock()
        init = zne_init(base_init)
        kwargs = {"Tolkien": "Silmarilion", "Moises": "Genesis"}

        # Without zne_strategy
        _ = init(self, options="options", **kwargs)
        base_init.assert_called_once_with(self, options="options", **kwargs)

        # With zne_strategy
        base_init.reset_mock()
        zne_strategy = ZNEStrategy()
        _ = init(self, options="options", zne_strategy=zne_strategy, **kwargs)
        base_init.assert_called_once_with(self, options="options", **kwargs)
        assert self.zne_strategy is zne_strategy


@mark.parametrize("init", [None, Ellipsis, True, 0, 1.0, 1j, "init"])
def test_type_error(init):
    with raises(TypeError):
        zne_init(init)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module for singleton type to represent unset args."""

from __future__ import annotations

from typing import Any


class UnsetType:
    """Singleton type to represent unset args."""

    __slots__ = ()

    def __new__(cls) -> UnsetType:
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance

    def __bool__(self) -> bool:
        return False

    def __eq__(self, __o: Any) -> bool:
        return self is __o

    def __repr__(self) -> str:
        return "UNSET"

    def __copy__(self) -> UnsetType:
        return self

    def __deepcopy__(self, memo: dict) -> UnsetType:
        return self


UNSET = UnsetType()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Grouping utils module."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from typing import Any


def group_elements(elements: Sequence, group_size: int) -> list[tuple]:
    """Group elements in iterable into tuples of a given size.

    Args:
        elements: A list of elements to be grouped
        group_size: The size of grouped tuples

    Returns:
        List of grouped tuples
    """
    return list(group_elements_gen(elements, group_size))


def group_elements_gen(elements: Sequence, group_size: int) -> Iterator[tuple]:
    """Generate groups of elements from iterable as tuples of a given size.

    Args:
        elements: A list of elements to be grouped
        group_size: The size of grouped tuples

    Yields:
        The next grouped tuple
    """
    if not isinstance(elements, Iterable):
        raise TypeError("Values argument must be iterable.")
    if not isinstance(group_size, int):
        raise TypeError("Size argument must be non-zero positive int.")
    if group_size < 1:
        raise ValueError("Size argument must be non-zero positive int.")
    elements = list(elements)
    while elements:
        yield tuple(elements[:group_size])
        del elements[:group_size]


def merge_dicts(dict_list: Sequence[dict]) -> dict:
    """Given a sequence of dictionaries merge them all into one.

    Args:
        dict_list: A sequence of dictuinaries.

    Returns:
        A dictionary resulting from merging all the input ones. In case of key collision,
        only the latest value will be preserved.
    """
    dictionary: dict = {}
    for dct in dict_list:
        dictionary.update(dct)
    return dictionary


def from_common_key(dict_list: Sequence[dict], key: Any) -> tuple:
    """Given a sequence of dictionaries and a common key extract all values.

    Args:
        dict_list: A sequence of dictuinaries

    Returns:
        A tuple with all values extracted from the input dictionaries. Whenever key is not
        present in a dictionary, ``None`` will be inserted instead of the corresponding value.
    """
    return tuple(d.get(key, None) for d in dict_list)


def squash_for_equal_elements(sequence: Sequence) -> Any:
    """Squash values in sequence to a single output if all are equal.

    Args:
        sequence: the sequence of values to collapse.

    Returns:
        The common value if all elements are equal, or the input sequence otherwise.
    """
    if not isinstance(sequence, Sequence):
        raise TypeError("Expected Sequence object, received {type(self.noise_amplifier)} instead.")
    if len(sequence) == 0:
        raise ValueError("Empty Sequence provided.")
    first = sequence[0]
    if all(element == first for element in sequence):
        return first
    return sequence



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Noise amplification strategy via circuit/circuit-inverse repetition."""


from qiskit.circuit import QuantumCircuit

from .folding_amplifier import FoldingAmplifier


class GlobalFoldingAmplifier(FoldingAmplifier):
    """Alternatingly composes the circuit and its inverse as many times as indicated by the
    ``noise_factor``.

    If ``noise_factor`` is not an odd integer, a sub part of the full circle is folded such that the
    resultant noise scaling :math:`\\lambda` equals :math:`\\lambda=(2n+1)+2s/d`, where :math:`d`
    are the number of layers of the original circuit, :math:`n` is the number of global circuit
    foldings, and :math:`s` is the number of gates that are sub-folded [1].

    The gates of the full circuit that are used for the sub-folding can be the first :math:`s`
    gates, the last :math:`s` gates or a random selection of :math:`s` gates. This is specified by
    ``sub_folding_option``.

    Note that all instance objects and their attributes are immutable by construction. If another
    strategy with different options needs to be utilized, a new instance of the class should be
    created.

    References:
        [1] T. Giurgica-Tiron et al. (2020).
            Digital zero noise extrapolation for quantum error mitigation.
            `<https://ieeexplore.ieee.org/document/9259940>`
    """

    def amplify_circuit_noise(self, circuit: QuantumCircuit, noise_factor: float) -> QuantumCircuit:
        self._validate_noise_factor(noise_factor)
        num_full_foldings, num_sub_foldings = self._compute_folding_nums(noise_factor, len(circuit))
        noisy_circuit = circuit.copy_empty_like()
        noisy_circuit = self._apply_full_folding(noisy_circuit, circuit, num_full_foldings)
        noisy_circuit = self._apply_sub_folding(noisy_circuit, circuit, num_sub_foldings)
        return noisy_circuit

    def _apply_full_folding(
        self, noisy_circuit: QuantumCircuit, original_circuit: QuantumCircuit, num_foldings: int
    ) -> QuantumCircuit:
        """Fully folds the original circuit a number of ``num_foldings`` times.

        Args:
            noisy_circuit: The noise amplified circuit to which the gates are added.
            original_circuit: The original circuit without foldings.
            num_foldings: Number of times the circuit should be folded.

        Returns:
            The noise amplified circuit.
        """
        noise_factor: int = self.folding_to_noise_factor(num_foldings)  # type: ignore
        for i in range(noise_factor):
            if i % 2 == 0:
                noisy_circuit.compose(original_circuit, inplace=True)
            else:
                noisy_circuit.compose(original_circuit.inverse(), inplace=True)
            noisy_circuit = self._apply_barrier(noisy_circuit)
        return noisy_circuit

    def _apply_sub_folding(
        self, noisy_circuit: QuantumCircuit, original_circuit: QuantumCircuit, num_foldings: int
    ) -> QuantumCircuit:
        """Folds a subset of gates of the original circuit.

        Args:
            noisy_circuit: The noise amplified circuit to which the gates are added.
            original_circuit: The original circuit without foldings.
            num_foldings: Number of gates to be folded.

        Returns:
            The noise amplified circuit.
        """
        if num_foldings == 0:
            return noisy_circuit
        sub_circuit: QuantumCircuit = self._get_sub_folding(original_circuit, num_foldings)
        noisy_circuit.compose(sub_circuit.inverse(), inplace=True)
        noisy_circuit.compose(sub_circuit, inplace=True)
        return noisy_circuit

    def _get_sub_folding(self, circuit: QuantumCircuit, size: int) -> QuantumCircuit:
        """Returns sub circuit to be folded.

        Args:
            circuit: The original circuit without foldings.
            size: Number of gates to be sub-folded.

        Returns:
            The sub circuit.
        """
        sub_circuit = circuit.copy_empty_like()
        if self._sub_folding_option == "from_first":
            sub_data = circuit.data[:size]
        elif self._sub_folding_option == "from_last":
            sub_data = circuit.data[-size:]
        else:
            instruction_idxs = sorted(self._rng.choice(len(circuit), size=size, replace=False))
            sub_data = [circuit.data[i] for i in instruction_idxs]
        for instruction, qargs, cargs in sub_data:
            # sub_circuit.barrier(qargs)  # TODO: avoid sub-folded gates simplification
            sub_circuit.append(instruction, qargs, cargs)
        return sub_circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


from test import NO_INTS

from numpy import allclose, random
from pytest import mark, raises

from zne.extrapolation import (
    CubicExtrapolator,
    LinearExtrapolator,
    PolynomialExtrapolator,
    QuadraticExtrapolator,
    QuarticExtrapolator,
)
from zne.utils.unset import UNSET

################################################################################
## AUXILIARY
################################################################################
ATOL: float = 1e-2
RTOL: float = 1e-5

MAX_DEGREE: int = 5


def extrapolate_zero_test_cases(max_degree):
    for degree in range(1, max_degree + 1):  # All degrees up to max
        min_points = degree + 1
        for coefficients in (
            [1 for _ in range(min_points)],
            [-1 for _ in range(min_points)],
            [c for c in range(min_points)],
            [1 + c for c in range(min_points)],
            [1 - c for c in range(min_points)],
        ):  # Different curves
            model = lambda x: PolynomialExtrapolator(degree)._model(x, *coefficients)
            for extra in range(5):  # Different number of data points
                x_data = [1 + x for x in range(min_points + extra)]
                y_data = [model(x) + random.normal(0, 1e-6) for x in x_data]
                sigma_y = [random.normal(0.1, 1e-4) for _ in y_data]
                expected = model(0)
                yield degree, x_data, y_data, sigma_y, expected


################################################################################
## EXTRAPOLATORS
################################################################################
class TestPolynomialExtrapolator:
    """Test polynomial extrapolator."""

    ################################################################################
    ## TESTS
    ################################################################################
    @mark.parametrize(
        "degree", cases := range(1, 5 + 1), ids=[f"degree = {degree!r}" for degree in cases]
    )
    def test_degree(self, degree):
        """Test degree."""
        extrapolator = PolynomialExtrapolator(degree=degree)
        assert isinstance(extrapolator.degree, int)
        assert extrapolator.degree == degree
        extrapolator = PolynomialExtrapolator(degree=str(degree))
        assert isinstance(extrapolator.degree, int)
        assert extrapolator.degree == degree
        extrapolator = PolynomialExtrapolator(degree=float(degree))
        assert isinstance(extrapolator.degree, int)
        assert extrapolator.degree == degree

    @mark.parametrize(
        "degree",
        cases := [d for d in NO_INTS if not isinstance(d, (str, float))],
        ids=[f"degree = {degree!r}" for degree in cases],
    )
    def test_degree_type_error(self, degree):
        """Test degree type error."""
        with raises(TypeError):
            PolynomialExtrapolator(degree)

    @mark.parametrize(
        "degree",
        cases := list(range(0, -5, -1)) + [str(float(d)) for d in range(1, 5 + 1)],
        ids=[f"degree = {degree!r}" for degree in cases],
    )
    def test_degree_value_error(self, degree):
        """Test degree value error."""
        with raises(ValueError):
            PolynomialExtrapolator(degree)

    @mark.parametrize("degree", cases := range(1, 5 + 1), ids=[f"{degree=}" for degree in cases])
    def test_min_points(self, degree):
        """Test min points."""
        extrapolator = PolynomialExtrapolator(degree=degree)
        assert extrapolator.min_points == degree + 1

    @mark.parametrize(
        "degree, x_data, y_data, sigma_y, expected",
        [*extrapolate_zero_test_cases(MAX_DEGREE)],
    )
    def test_extrapolate_zero(self, degree, x_data, y_data, sigma_y, expected):
        """Test extrapolate zero."""
        extrapolator = PolynomialExtrapolator(degree)
        value, std_error, metadata = extrapolator.extrapolate_zero(x_data, y_data, sigma_y=sigma_y)
        assert allclose(value, expected, atol=ATOL, rtol=RTOL)
        assert isinstance(std_error, float)  # TODO: test value
        for key in ["coefficients", "covariance_matrix", "residuals", "R2"]:
            assert metadata.get(key, UNSET) is not UNSET  # TODO: test values


################################################################################
## FACADES
################################################################################
class TestFacades:
    """Test polynomial extrapolator facades."""

    @mark.parametrize(
        "cls, configs",
        [
            (LinearExtrapolator, {"degree": 1}),
            (QuadraticExtrapolator, {"degree": 2}),
            (CubicExtrapolator, {"degree": 3}),
            (QuarticExtrapolator, {"degree": 4}),
        ],
    )
    def test_facades(self, cls, configs):
        """Test polynomial extrapolator facades."""
        # Note: using `strategy` decorator functionality
        assert cls() == PolynomialExtrapolator(**configs)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Multi-exponential ordinary-least-squares (OLS) extrapolator.

Theoretical results [1] point at the multi-exponential model being the
most suitable for performing zero-noise extrapolation; nonetheless, the
nature of this regression model, where all the exponential terms are
identical, make it unstable to fit in practice [2] for more than one
exponential term. For the sake of generality, this class provides an
extrapolator for the multi-exponential regression model OLS fitted:
`shift + sum(amplitude[i] * exp(-rate[i] * x) for i in range(num_terms))`

Notice that a multi-exponential decay always tends towards zero, however,
a general observable will tend towards the average of its eigenvalues
(i.e. trace over the number of dimensions) for a state closer and closer
to the complete mixed state (i.e. as noise increases). Generally speaking,
such average will not be zero, and therefore, the multi-exponential decay
will not be a valid model in all scenarios. Nonetheless, by expressing
such general observable as a sum Pauli operators, it is easy to notice
that the only non-traceless element will be the one associated to the
identity operator (i.e. other Paulis are always traceless), hence
determining the expectation value that the general observable will tend
towards. Since the expectation value of the identity operator is trivial
to obtain, by virtue of subtracting such term from the sum of Paulis, the
remaining observable will naturally tend towards zero just as the
multi-exponential model assumes. However, coherent noise can make the
expected asymptote to drift. For that reason, this regression model
includes one extra shift coefficient in the form of a constant factor
added to the exponential terms globally.

Extrapolating constant noise profiles will generally fail due to a poor
convergence of the model. There are several reasons for this:
    1. Assuming all decay rates are fitted to zero, there is ambiguity in
    how the constant factor should be distributed across the shift
    coefficient and the amplitudes.
    2. Even with non-zero amplitude coefficients, one could fit
    arbitrarily many data points on the constant by virtue of boosting
    the decay rates; as this will make the contributions from the
    exponential terms arbitrarily close to zero for value of `x!=0`.

There are two main scenarios in the context of ZNE when this is relevant:
    1. If noise saturation has been reached for all input data points;
    in which case no relevant information can be extracted and
    extrapolation is therefore meaningless.
    2. If the desired mitigated value is close to the decayed value;
    in which case the instability will prevent from achieving meaningful
    error mitigation by providing either bad accuracy, precision or both.

Solving for this phenomena is beyond the scope of this class, but users
may consider techniques such as:
    1. After extrapolation: Comparing with other extrapolation methods
    (e.g. linear) and choosing the result based on metrics such as the std
    error of the extrapolated value, or goodness of fit.
    2. Before extrapolation: trying to predict whether the profile
    corresponds to a constant before fitting by pre-analyzing the data
    points, and manually assigning the coefficient values if so.

Notice however that, a priori, there is no way to identify which of the
possible scenarios we may be operating under. Therefore, one always has to
be wary of constant noise profiles regardless of the extrapolator;
as sometimes they will carry no information about the ideal value (i.e.
scenario #1), and sometimes they will portray relevant information and
be meaningful (i.e. scenario #2).

References:
    [1] Cai, Zhenyu. "Multi-exponential error extrapolation and combining
    error mitigation techniques for NISQ applications." npj Quantum
    Information 7 (2020): 1-12.
    [2] Bi, C., Fishbein, K., Bouhrara, M. et al. "Stabilization of
    parameter estimates from multiexponential decay through extension
    into higher dimensions." Sci Rep 12, 5773 (2022).
"""

from __future__ import annotations

from numpy import array, exp, inf, ndarray, ones, sqrt
from scipy.optimize import curve_fit

from zne.utils.grouping import group_elements_gen

from .extrapolator import OLSExtrapolator, ReckoningResult


################################################################################
## GENERAL
################################################################################
class MultiExponentialExtrapolator(OLSExtrapolator):
    """Multi-exponential ordinary-least-squares (OLS) extrapolator.

    Args:
        num_terms: The number of exponential terms `amplitude * exp(-rate * x)`
            added together in the regression model.
    """

    def __init__(self, num_terms: int = 1):  # pylint: disable=super-init-not-called
        self._set_num_terms(num_terms)

    ################################################################################
    ## PROPERTIES
    ################################################################################
    @property
    def num_terms(self) -> int:
        """The number of exponential terms added in the regression model."""
        return self._num_terms

    def _set_num_terms(self, num_terms: int) -> None:
        """Number of terms setter."""
        num_terms = int(num_terms)
        if num_terms < 1:
            raise ValueError("Number of exponential terms must be at least 1.")
        self._num_terms: int = num_terms

    ################################################################################
    ## IMPLEMENTATION
    ################################################################################
    @property
    def min_points(self) -> int:
        return self.num_terms * 2 + 1

    # pylint: disable=duplicate-code
    def _extrapolate_zero(
        self,
        x_data: tuple[float, ...],
        y_data: tuple[float, ...],
        sigma_x: tuple[float, ...],  # pylint: disable=unused-argument
        sigma_y: tuple[float, ...],
    ) -> ReckoningResult:
        coefficients, covariance_matrix = curve_fit(
            self._model,
            x_data,
            y_data,
            sigma=self._compute_sigma(y_data, sigma_y),
            absolute_sigma=True,
            p0=[2 ** (-i) for i in range(self.num_terms * 2 + 1)],
            bounds=([-inf] + [-inf, 0] * self.num_terms, inf),  # Note: only decay considered
            max_nfev=None,
        )
        value = self._model(0, *coefficients)
        entries = ones(self.num_terms * 2 + 1)
        entries[2::2] = 0  # Note: decay rate coefficients not relevant for x=0
        variance = entries @ covariance_matrix @ entries
        std_error = sqrt(variance)
        metadata = self._build_metadata(
            array(x_data),
            array(y_data),
            coefficients,
            covariance_matrix,
        )
        return ReckoningResult(value.tolist(), std_error.tolist(), metadata)

    def _model(self, x, *coefficients) -> ndarray:  # pylint: disable=invalid-name
        """Exponential regression model for curve fitting."""
        x = array(x)
        shift = coefficients[0]
        exp_coefficients = coefficients[1:]
        exponential = sum(
            amplitude * exp(-rate * x)  # Note: decay for positive values of `rate`
            for amplitude, rate in group_elements_gen(exp_coefficients, group_size=2)
        )
        return exponential + shift


################################################################################
## FACADES
################################################################################
class MonoExponentialExtrapolator(MultiExponentialExtrapolator):
    """Mono-exponential ordinary-least-squares (OLS) extrapolator."""

    def __init__(self):
        super().__init__(num_terms=1)


class BiExponentialExtrapolator(MultiExponentialExtrapolator):
    """Bi-exponential ordinary-least-squares (OLS) extrapolator.

    Note: The convergence of this model is unstable. Proceed with caution.
    """

    def __init__(self):
        super().__init__(num_terms=2)


class ExponentialExtrapolator(MonoExponentialExtrapolator):
    """Exponential ordinary-least-squares (OLS) extrapolator."""



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interface for extrapolation strategies."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import namedtuple
from collections.abc import Sequence

from numpy import array
from numpy import float_ as npfloat
from numpy import isclose, mean, ndarray, ones

from zne.types import Metadata
from zne.utils.strategy import strategy
from zne.utils.typing import isreal, normalize_array

# TODO: import from staged_primitives
ReckoningResult = namedtuple("ReckoningResult", ("value", "std_error", "metadata"))


################################################################################
## EXTRAPOLATOR
################################################################################
@strategy
class Extrapolator(ABC):
    """Interface for extrapolation strategies."""

    @property
    @abstractmethod
    def min_points(self) -> int:
        """The minimum number of data points required to fit the regression model."""
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _extrapolate_zero(
        self,
        x_data: tuple[float, ...],
        y_data: tuple[float, ...],
        sigma_x: tuple[float, ...],
        sigma_y: tuple[float, ...],
    ) -> ReckoningResult:
        """Core functionality for `extrapolate_zero` after input validation."""
        raise NotImplementedError  # pragma: no cover

    ################################################################################
    ## API
    ################################################################################
    def extrapolate_zero(
        self,
        x_data: Sequence[float],
        y_data: Sequence[float],
        sigma_x: Sequence[float] | None = None,
        sigma_y: Sequence[float] | None = None,
    ) -> ReckoningResult:
        """Extrapolate to zero by fitting a regression model to the provided data.

        Args:
            x_data: A sequence of X values for the data points to fit.
            y_data: A sequence of Y values for the data points to fit.
            sigma_x: A sequence of std errors along the X axis for the data points to fit.
                If `None`, ones of `x_data` size is assumed.
            sigma_y: A sequence of std errors along the Y axis for the data points to fit.
                If `None`, ones of `y_data` size is assumed.

        Returns:
            A ReckoningResult namedtuple object holding the extrapolated value, std error,
            and metadata.
        """
        x_data = self._validate_data(x_data)
        y_data = self._validate_data(y_data)
        sigma_x = self._validate_sigma(sigma_x, default_size=len(x_data))
        sigma_y = self._validate_sigma(sigma_y, default_size=len(y_data))
        self._cross_validate_data_sigma(x_data, y_data, sigma_x, sigma_y)
        self._validate_min_points(x_data, y_data)
        return self._extrapolate_zero(x_data, y_data, sigma_x, sigma_y)

    ################################################################################
    ## VALIDATION
    ################################################################################
    def _validate_data(self, data: Sequence[float]) -> tuple[float, ...]:
        """Validates data for the regression model.

        Args:
            data: A sequence of values for the regression model.

        Returns:
            Validated data normalized.
        """
        if not isinstance(data, Sequence) or not all(isreal(d) for d in data):
            raise TypeError(f"Invalid data {data}, expected sequence of floats.")
        return tuple(data)

    def _validate_sigma(
        self, sigma: Sequence[float] | None, default_size: int
    ) -> tuple[float, ...]:
        """Validates sigma for the regression model data.

        Args:
            sigma: A sequence of values for the std error in the regression model data.
            default_size: Default size of the sigma to be returned if `None`.

        Returns:
            Validated sigma normalized. If `None`, ones of default size is returned.
        """
        if sigma is None:
            sigma = [1] * default_size  # Note: zeros mean that the data is exact
        return self._validate_data(sigma)

    def _cross_validate_data_sigma(
        self,
        x_data: tuple[float, ...],
        y_data: tuple[float, ...],
        sigma_x: tuple[float, ...],
        sigma_y: tuple[float, ...],
    ) -> None:
        """Cross-validate data and sigmas."""
        sizes = {len(x_data), len(y_data), len(sigma_x), len(sigma_y)}
        if len(sizes) != 1:
            raise ValueError(
                "Invalid data, all inputs should be of the same size: "
                f"{len(x_data) = }, {len(y_data) = }, "  # noqa: E251,E202
                f"{len(sigma_x) = }, {len(sigma_y) = }."  # noqa: E251,E202
            )

    def _validate_min_points(
        self,
        x_data: tuple[float, ...],
        y_data: tuple[float, ...],  # pylint: disable=unused-argument
    ) -> None:
        """Validate that min points threshold is achieved."""
        num_points = len(set(x_data))  # TODO: equal up to tolerance
        if num_points < self.min_points:
            raise ValueError(
                f"Insufficient number of distinct data points provided ({num_points}), "
                f"at least {self.min_points} needed."
            )


################################################################################
## OLS EXTRAPOLATOR
################################################################################
class OLSExtrapolator(Extrapolator):
    """Interface for ordinary-least-squares (OLS) extrapolation strategies."""

    @abstractmethod
    def _model(self, x, *coefficients) -> ndarray:  # pylint: disable=invalid-name
        """Regression model for curve fitting."""
        raise NotImplementedError  # pragma: no cover

    ################################################################################
    ## AUXILIARY
    ################################################################################
    def _compute_sigma(
        self,
        y_data: tuple[float, ...],
        sigma_y: tuple[float, ...],
    ) -> ndarray:
        """Compute sensible sigma values for curve fitting.

        This implementation bypasses zero effective variance which would
        lead to numerical errors in the curve fitting procedure.
        """
        values = array(y_data)
        errors = array(sigma_y)
        relative_errors = errors / values
        return ones(errors.shape) if any(isclose(relative_errors, 0)) else errors

    def _build_metadata(
        self,
        x_data: ndarray,
        y_data: ndarray,
        coefficients: ndarray,
        covariance_matrix: ndarray,
    ) -> Metadata:
        """Build regression metadata."""
        residuals = y_data - self._model(x_data, *coefficients)
        r_squared = self._r_squared(y_data, residuals)
        return {
            "coefficients": normalize_array(coefficients),
            "covariance_matrix": normalize_array(covariance_matrix),
            "residuals": normalize_array(residuals),
            "R2": normalize_array(r_squared),
        }

    @staticmethod
    def _r_squared(y_data: ndarray, residuals: ndarray) -> npfloat:
        """Compute R-squared (i.e. coefficient of determination)."""
        y_diff = y_data - mean(y_data)
        TSS = y_diff @ y_diff  # pylint: disable=invalid-name
        RSS = residuals @ residuals  # pylint: disable=invalid-name
        return 1 - RSS / TSS



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from itertools import product
from unittest.mock import Mock, patch

from numpy.random import default_rng
from pytest import approx, fixture, mark, raises
from qiskit.circuit import Barrier, CircuitInstruction, ParameterVector, QuantumCircuit
from qiskit.circuit.random import random_circuit
from qiskit.quantum_info.operators import Operator
from qiskit.transpiler.passes import RemoveBarriers

from zne.noise_amplification.folding_amplifier import GlobalFoldingAmplifier

MOCK_TARGET_PATH = (
    "zne.noise_amplification.folding_amplifier.global_folding_amplifier.GlobalFoldingAmplifier"
)


################################################################################
## MODULE FIXTURES
################################################################################
@fixture(scope="module")
def noise_amplifier():
    return GlobalFoldingAmplifier()


@fixture(scope="module")
def get_sub_circuit():
    def factory_method(circuit, sub_folding_option, num_foldings=1, random_seed=0):
        noise_amplifier = GlobalFoldingAmplifier(
            sub_folding_option=sub_folding_option, random_seed=random_seed
        )
        return noise_amplifier._get_sub_folding(circuit, num_foldings)

    return factory_method


@fixture(scope="module")
def patch_amplifier_with_mock():
    def factory_method(method_name=None, **kwargs):
        method_name = method_name or ""
        return patch(MOCK_TARGET_PATH + method_name, **kwargs)

    return factory_method


@fixture(scope="module")
def patch_amplifier_with_multiple_mocks():
    def factory_method(**kwargs):
        return patch.multiple(MOCK_TARGET_PATH, **kwargs)

    return factory_method


################################################################################
## PUBLIC METHODS TESTS
################################################################################
def test_amplify_circuit_noise(
    noise_amplifier, circuit, get_random_circuit, patch_amplifier_with_multiple_mocks
):
    # TODO: rewrite test with class refactor
    mocks = {
        "_validate_noise_factor": Mock(),
        "_compute_folding_nums": Mock(return_value=(1, 2)),
        "_apply_full_folding": Mock(return_value=get_random_circuit(2)),
        "_apply_sub_folding": Mock(return_value=get_random_circuit(3)),
    }
    circuit_mock_target = (
        "zne.noise_amplification.folding_amplifier.global_folding_amplifier"
        ".QuantumCircuit.copy_empty_like"
    )
    with patch_amplifier_with_multiple_mocks(**mocks):
        with patch(circuit_mock_target, return_value=get_random_circuit(1)) as _:
            _ = noise_amplifier.amplify_circuit_noise(circuit, 3)
    mocks["_validate_noise_factor"].assert_called_once_with(3)
    # circuit_copy_mock.assert_called_once_with()
    mocks["_compute_folding_nums"].assert_called_once_with(3, len(circuit))
    # mocks["_apply_full_folding"].assert_called_once_with(get_random_circuit(1), circuit, 1)
    # mocks["_apply_sub_folding"].assert_called_once_with(get_random_circuit(2), circuit, 2)
    # assert noisy_circuit == get_random_circuit(3)


################################################################################
## PRIVATE METHODS TESTS
################################################################################
@mark.parametrize(
    "circuit_seed, num_foldings",
    cases := tuple(
        product(
            [0, 5, 66, 1081, 1082],
            [0, 1, 2, 3],
        )
    ),
    ids=[f"QuantumCircuit<{cs}>-{nf}" for cs, nf in cases],
)
def test_apply_full_folding(noise_amplifier, get_random_circuit, circuit_seed, num_foldings):
    circuit = get_random_circuit(circuit_seed)
    noisy_circuit = circuit.copy_empty_like()
    noisy_circuit = noise_amplifier._apply_full_folding(noisy_circuit, circuit, num_foldings)
    noisy_operation = (operation for operation in noisy_circuit)
    num_gates = 0
    for i in range(2 * num_foldings + 1):
        if i % 2 == 0:
            for original_operation in circuit:
                assert next(noisy_operation) == original_operation
                num_gates += 1
        else:
            for original_operation in circuit.inverse():
                assert next(noisy_operation) == original_operation
                num_gates += 1
        instruction, qargs, cargs = next(noisy_operation)
        assert instruction == Barrier(circuit.num_qubits)
        assert qargs == noisy_circuit.qubits
        assert cargs == []
    assert 2 * num_foldings + 1 == num_gates / len(circuit)


@mark.parametrize(
    "noisy_circuit",
    cases := [
        random_circuit(2, 3, seed=0),
        random_circuit(2, 6, seed=5),
    ],
    ids=[f"{type(c).__name__}<{len(c)}>" for c in cases],
)
def test_apply_sub_folding_wo_folding(noise_amplifier, circuit, noisy_circuit):
    assert noisy_circuit == noise_amplifier._apply_sub_folding(noisy_circuit, circuit, 0)


@mark.parametrize(
    "sub_circuit, num_foldings",
    cases := tuple(
        zip(
            [
                random_circuit(2, 3, seed=0),
                random_circuit(2, 6, seed=5),
            ],
            [1, 3],
        )
    ),
    ids=[f"{type(c).__name__}<{len(c)}>-{nf}" for c, nf in cases],
)
def test_apply_sub_folding(
    noise_amplifier,
    circuit,
    get_random_circuit,
    patch_amplifier_with_mock,
    sub_circuit,
    num_foldings,
):
    noisy_circuit_in = get_random_circuit(3)
    with patch_amplifier_with_mock(
        method_name="._get_sub_folding", return_value=sub_circuit
    ) as mock:
        noisy_circuit_out = noise_amplifier._apply_sub_folding(
            noisy_circuit_in.copy(), circuit, num_foldings
        )
        mock.assert_called_once_with(circuit, num_foldings)
    original_circuit_length = len(noisy_circuit_in)
    assert noisy_circuit_out.data[:original_circuit_length] == noisy_circuit_in.data
    extended_circuit_length = original_circuit_length + len(sub_circuit)
    noisy_circuit_slice = noisy_circuit_out.data[original_circuit_length:extended_circuit_length]
    assert noisy_circuit_slice == sub_circuit.inverse().data
    assert noisy_circuit_out.data[extended_circuit_length:] == sub_circuit.data


@mark.parametrize(
    "circuit, num_foldings",
    cases := tuple(
        product(
            [
                random_circuit(2, 3, seed=0),
                random_circuit(2, 3, seed=5),
                random_circuit(2, 6, seed=66),
                random_circuit(2, 6, seed=1081),
            ],
            [1, 2, 3],
        )
    ),
    ids=[f"{type(c).__name__}<{len(c)}>-{nf}" for c, nf in cases],
)
class TestGetSubFolding:
    def test_sub_folding_from_last(self, get_sub_circuit, circuit, num_foldings):
        sub_circuit = get_sub_circuit(circuit, "from_last", num_foldings)
        sub_data = (operation for operation in sub_circuit)
        original_data = (operation for operation in circuit[-num_foldings:])
        for _ in range(num_foldings):
            o_instruction, o_qargs, o_cargs = next(original_data)
            # instruction, qargs, cargs = next(sub_data)
            # assert instruction == Barrier(len(o_qargs))
            # assert qargs == o_qargs
            # assert cargs == []
            instruction, qargs, cargs = next(sub_data)
            assert instruction == o_instruction
            assert qargs == o_qargs
            assert cargs == o_cargs
        with raises(StopIteration):
            next(original_data)
        with raises(StopIteration):
            next(sub_data)

    def test_sub_folding_from_first(self, get_sub_circuit, circuit, num_foldings):
        sub_circuit = get_sub_circuit(circuit, "from_first", num_foldings)
        sub_data = (operation for operation in sub_circuit)
        original_data = (operation for operation in circuit[:num_foldings])
        for _ in range(num_foldings):
            o_instruction, o_qargs, o_cargs = next(original_data)
            # instruction, qargs, cargs = next(sub_data)
            # assert instruction == Barrier(len(o_qargs))
            # assert qargs == o_qargs
            # assert cargs == []
            instruction, qargs, cargs = next(sub_data)
            assert instruction == o_instruction
            assert qargs == o_qargs
            assert cargs == o_cargs
        with raises(StopIteration):
            next(original_data)
        with raises(StopIteration):
            next(sub_data)

    @mark.parametrize("seed", cases := [1, 5, 66], ids=[f"{s}" for s in cases])
    def test_get_sub_folding_random(self, get_sub_circuit, circuit, num_foldings, seed):
        sub_circuit = get_sub_circuit(circuit, "random", num_foldings, random_seed=seed)
        sub_data = (operation for operation in sub_circuit)
        idxs = sorted(default_rng(seed).choice(len(circuit), size=num_foldings, replace=False))
        original_data = (circuit.data[i] for i in idxs)
        for _ in range(num_foldings):
            o_instruction, o_qargs, o_cargs = next(original_data)
            # instruction, qargs, cargs = next(sub_data)
            # assert instruction == Barrier(len(o_qargs))
            # assert qargs == o_qargs
            # assert cargs == []
            instruction, qargs, cargs = next(sub_data)
            assert instruction == o_instruction
            assert qargs == o_qargs
            assert cargs == o_cargs
        with raises(StopIteration):
            next(original_data)
        with raises(StopIteration):
            next(sub_data)


class TestParametrizedCircuits:
    @fixture(scope="class")
    def parametrized_circuit(self):
        circuit = QuantumCircuit(2)
        params = ParameterVector("P", 4)
        circuit.ry(params[0], 0)
        circuit.rz(params[0], 1)
        circuit.crx(params[1], 0, 1)
        circuit.rz(params[2], 0)
        circuit.cu(params[3], params[0], params[1], params[0], 1, 0)
        circuit.ry(params[3], 1)
        circuit.crx(params[3], 0, 1)
        return circuit

    @fixture(scope="function")
    def noisy_circuit(self, noise_amplifier, parametrized_circuit):
        noisy_circuit = parametrized_circuit.copy_empty_like()
        noisy_circuit = noise_amplifier._apply_full_folding(noisy_circuit, parametrized_circuit, 1)
        return noisy_circuit

    @mark.parametrize(
        "sub_folding_option, sub_folding_index",
        cases := tuple(
            zip(
                ["from_last", "from_first", "random"],
                [-1, 0, 3],
            )
        ),
        ids=["from_last", "from_first", "random"],
    )
    def test_get_sub_folding_with_parameters(
        self,
        get_sub_circuit,
        parametrized_circuit,
        sub_folding_option,
        sub_folding_index,
    ):
        sub_circuit = get_sub_circuit(
            parametrized_circuit, sub_folding_option, num_foldings=1, random_seed=1
        )
        assert sub_circuit.data[0] == parametrized_circuit.data[sub_folding_index]
        assert len(sub_circuit._parameter_table) == 1

    def test_apply_full_folding_circuit_parameters(self, parametrized_circuit, noisy_circuit):
        noisy_operation = (operation for operation in noisy_circuit)
        barrier = CircuitInstruction(Barrier(noisy_circuit.num_qubits), noisy_circuit.qubits, [])
        for circuit in [parametrized_circuit, parametrized_circuit.inverse(), parametrized_circuit]:
            for original_instruction, _, _ in circuit:
                instruction, _, _ = next(noisy_operation)
                assert instruction.params == original_instruction.params
            assert next(noisy_operation) == barrier


################################################################################
## INTEGRATION TESTS
################################################################################
@mark.parametrize(
    "circuit, noise_factor, sub_folding_option",
    cases := tuple(
        product(
            [
                random_circuit(5, 3, seed=0),
                random_circuit(5, 3, seed=5),
                random_circuit(5, 6, seed=66),
                random_circuit(5, 6, seed=1081),
            ],
            [1.5, 3.3, 17 / 3],
            ["from_last", "from_first", "random"],
        )
    ),
    ids=[f"{type(c).__name__}<{len(c)}>-{nf}-{sfo}" for c, nf, sfo in cases],
)
class TestIntegration:
    @mark.filterwarnings("ignore::UserWarning")
    def test_noise_factor_of_folded_circuit(self, circuit, noise_factor, sub_folding_option):
        noise_amplifier = GlobalFoldingAmplifier(sub_folding_option=sub_folding_option)
        noisy_circuit = noise_amplifier.amplify_circuit_noise(circuit.copy(), noise_factor)
        noisy_circuit = RemoveBarriers()(noisy_circuit)
        num_instructions = len(circuit)
        num_foldings = round(num_instructions * (noise_factor - 1) / 2.0)
        closest_noise_factor = 2 * num_foldings / num_instructions + 1
        assert len(noisy_circuit) / len(circuit) == approx(closest_noise_factor)
        assert Operator(noisy_circuit).equiv(Operator(circuit))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from collections.abc import Iterable

from zne import __version__


def test_version():
    assert __version__ == "1.3.0"


################################################################################
## DEFINITIONS
################################################################################
TYPES = [
    INT := 0,
    FLOAT := 0.0,
    NAN := float("NaN"),
    INF := float("Inf"),
    MINF := float("-Inf"),
    COMPLEX := complex(0, 0),
    STR := "0",
    BOOL := True,
    NONE := None,
    LIST := [0],
    TUPLE := (0,),
    DICT := {0: 0},
]
NO_INTS = [t for t in TYPES if not isinstance(t, int)]
NO_NONE = [t for t in TYPES if t is not None]
NO_INTS_NONE = [t for t in NO_INTS if t is not None]
NO_REAL = [t for t in NO_INTS if not isinstance(t, float)]
NO_NUM = [t for t in NO_REAL if not isinstance(t, complex)]
ITERS = [t for t in TYPES if isinstance(t, Iterable)]
NO_ITERS = [t for t in TYPES if not isinstance(t, Iterable)]
NO_ITERS_NONE = [t for t in NO_ITERS if t is not None]


################################################################################
## ALL
################################################################################
__all__ = [
    "TYPES",
    "NO_INTS",
    "NO_INTS_NONE",
    "NO_REAL",
    "NO_NUM",
    "ITERS",
    "NO_ITERS",
]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from unittest.mock import Mock

from pytest import mark, raises

from zne.utils.strategy import (
    _BaseStrategy,
    _closest_common_ancestor,
    _FrozenStrategy,
    _infer_init_namespace,
    _infer_settings_namespace,
    _pack_init_args,
    _shared_strategy_ancestor,
    strategy,
)


################################################################################
## TEST CASES
################################################################################
def general_ancestry():
    """Generate test cases for general common ancestry.

    Returns:
        A three-tuple holding the two classes to compare and the common ancestor class.
    """
    cls = type("cls", (), {})
    clz = type("clz", (), {})
    A = type("A", (cls,), {})
    B = type("B", (cls,), {})
    AA = type("AA", (A,), {})
    AB = type("AB", (A,), {})
    yield cls, int, object
    yield cls, clz, object
    yield cls, A, cls
    yield A, B, cls
    yield AA, B, cls
    yield AA, AB, A


def strategy_ancestry():
    """Generate test cases for strategy common ancestry.

    Returns:
        A three-tuple holding the two classes to compare and the common strategy ancestor.
    """
    # TODO: _FrozenStrategy
    NonStrategy = type("NonStrategy", (), {})
    cls = type("cls", (_BaseStrategy,), {})
    clz = type("clz", (_BaseStrategy,), {})
    A = type("A", (cls,), {})
    B = type("B", (cls,), {})
    AA = type("AA", (A,), {})
    AB = type("AB", (A,), {})
    yield cls, int, None
    yield cls, NonStrategy, None
    yield cls, clz, None
    yield cls, A, cls
    yield A, B, cls
    yield AA, B, cls
    yield AA, AB, A


def strategy_facades():
    """Generate test cases for strategy facades.

    Returns:
        A three-tuple holding the two classes to compare and whether the former is
        a facade of the latter.
    """
    NonStrategy = type("NonStrategy", (), {"__init__": lambda self, a, b: None})
    cls = type("cls", (_BaseStrategy,), {"__init__": lambda self, a, b: None})
    clz = type("clz", (_BaseStrategy,), {"__init__": lambda self, a, b: None})
    subcls = type("subcls", (cls,), {"__init__": lambda self, a, b: None})
    c_setting = type("c_setting", (cls,), {"__init__": lambda self, a, b, c: None})
    c_arg = type("c_arg", (cls,), {"__init__": lambda self, a, b, _c: None})
    A = type("A", (cls,), {"__init__": lambda self, a: None})
    B = type("B", (cls,), {"__init__": lambda self, b: None})
    C = type("C", (cls,), {"__init__": lambda self, c: None})
    _C = type("_C", (cls,), {"__init__": lambda self, _c: None})
    AA = type("AA", (A,), {"__init__": lambda self: None})
    AB = type("AB", (A,), {"__init__": lambda self, a, b: None})
    ABA = type("ABA", (AB,), {"__init__": lambda self, a: None})
    CA = type("AA", (C,), {"__init__": lambda self: None})
    yield cls, NonStrategy, False  # Non-strategy
    yield clz, cls, False  # Different strategies
    yield cls, cls, True  # Self facade
    yield subcls, cls, True  # Same settings facade
    yield c_setting, cls, False  # New setting
    yield c_arg, cls, True  # New excluded
    yield A, cls, True  # Simple facade
    yield B, cls, True  # Other facade
    yield C, cls, False  # Generalization
    yield _C, cls, True  # Excluded arg
    yield AA, cls, True  # Sub-facade
    yield CA, cls, False  # Intermediate generalization
    yield ABA, cls, False  # Local generalization


################################################################################
## CLASS DECORATOR
################################################################################
class TestClassDecorator:
    """Test strategy class decorator."""

    def test_new(self):
        """Test new."""
        decorator = strategy()
        assert isinstance(decorator, strategy)
        cls = type("cls", (), {})
        decorator = strategy(cls)
        assert not isinstance(decorator, strategy)
        assert issubclass(decorator, cls)
        assert issubclass(decorator, _BaseStrategy)
        assert not issubclass(decorator, _FrozenStrategy)
        decorator = strategy(cls, frozen=True)
        assert not isinstance(decorator, strategy)
        assert issubclass(decorator, cls)
        assert issubclass(decorator, _BaseStrategy)
        assert issubclass(decorator, _FrozenStrategy)

    def test_init(self):
        """Test init."""
        decorator = strategy()
        assert decorator.frozen is False
        decorator = strategy(frozen=True)
        assert decorator.frozen is True
        decorator = strategy(frozen=False)
        assert decorator.frozen is False

    def test_call(self):
        """Test call."""
        decorator = strategy()
        ## PLAIN STRATEGY
        cls = type("cls", (), {})
        strategy_class = decorator(cls)
        assert issubclass(strategy_class, cls)
        assert issubclass(strategy_class, _BaseStrategy)
        assert not issubclass(strategy_class, _FrozenStrategy)
        assert strategy_class.__init__ is not object.__init__
        strategy_class = decorator(cls, frozen=True)
        assert issubclass(strategy_class, cls)
        assert issubclass(strategy_class, _BaseStrategy)
        assert issubclass(strategy_class, _FrozenStrategy)
        assert strategy_class.__init__ is not object.__init__
        ## NAME
        name = "NAME"
        cls = type(name, (), {})
        assert decorator(cls).__name__ is name
        ## OVERRIDING MODULE
        module = "MODULE"
        cls = type("cls", (), {"__module__": module})
        assert decorator(cls).__module__ is module
        # OVERRIDING ANNOTATIONS
        annotations = {"annotation": ...}
        cls = type("cls", (), {"__annotations__": annotations})
        assert decorator(cls).__annotations__ is annotations
        ## OVERRIDING DOC
        doc = "DOCSTRING"
        cls = type("cls", (), {"__doc__": doc})
        assert decorator(cls).__doc__ is doc
        ## OVERRIDING INIT
        init = lambda self: None
        cls = type("cls", (), {"__init__": init})
        assert decorator(cls).__init__ is init
        ## OVERRIDING REPR
        repr = lambda self: "STRATEGY-REPR"
        cls = type("cls", (), {"__repr__": repr})
        assert decorator(cls).__repr__ is repr
        ## OVERRIDING EQ
        eq = lambda self, other: self is other
        cls = type("cls", (), {"__eq__": eq})
        assert decorator(cls).__eq__ is eq
        ## EXTENDING INIT SUBCLASS
        super_call = Mock()
        cls = type("cls", (), {"__init_subclass__": super_call})
        _ = decorator(cls)
        super_call.assert_called_once()
        ## EXTENDING NEW
        super_call = Mock()
        cls = type("cls", (), {"__new__": super_call})
        _ = decorator(cls)()
        super_call.assert_called_once()
        ## EXTENDING GETATTR
        super_call = Mock()
        cls = type("cls", (), {"__getattr__": super_call})
        getattr(decorator(cls)(), "attr")
        super_call.assert_called_once()

    def test_frozen(self):
        """Test frozen property."""
        decorator = strategy()
        assert decorator.frozen is False
        decorator.frozen = True
        assert decorator.frozen is True
        decorator.frozen = False
        assert decorator.frozen is False
        decorator.frozen = 1
        assert decorator.frozen is True
        decorator.frozen = {}
        assert decorator.frozen is False


################################################################################
## AUXILIARY
################################################################################
@mark.parametrize(
    "init, kwargs",
    [
        (lambda self: None, {}),
        (lambda self, a: None, {"a": 0}),
        (lambda self, a, b: None, {"a": 0, "b": 1}),
        (lambda self, a, b, c: None, {"a": 0, "b": 1, "c": 2}),
    ],
)
def test_pack_init_args(init, kwargs):
    """Test init args packing logic."""
    cls = type("cls", (), {"__init__": init})
    obj = cls(**kwargs)
    assert _pack_init_args(obj, **kwargs) == kwargs  # keyword
    assert _pack_init_args(obj, *kwargs.values()) == kwargs  # positional
    if len(kwargs) > 1:
        pos = tuple(v for v in tuple(kwargs.values())[: len(kwargs) // 2])
        kw = {k: v for k, v in tuple(kwargs.items())[len(kwargs) // 2 :]}
        assert _pack_init_args(obj, *pos, **kw) == kwargs  # positional + keyword


@mark.parametrize(
    "init, namespace",
    [
        (lambda self: None, ()),
        (lambda self, a: None, ("a",)),
        (lambda self, a, b: None, ("a", "b")),
        (lambda self, a, b, c: None, ("a", "b", "c")),
    ],
)
def test_infer_init_namesapce(init, namespace):
    """Test init namespace inferance."""
    cls = type("cls", (), {"__init__": init})
    assert _infer_init_namespace(cls) == namespace


@mark.parametrize(
    "init, settings",
    [
        (object.__init__, ()),
        (lambda self: None, ()),
        (lambda self, a: None, ("a",)),
        (lambda self, _a: None, ()),
        (lambda self, a, b: None, ("a", "b")),
        (lambda self, a, _b: None, ("a",)),
        (lambda self, _a, b: None, ("b",)),
        (lambda self, _a, _b: None, ()),
        (lambda self, a, _b, c: None, ("a", "c")),
    ],
)
def test_infer_settings_namesapce(init, settings):
    """Test settings namespace inferance."""
    cls = type("cls", (), {"__init__": init})
    assert _infer_settings_namespace(cls) == settings


@mark.parametrize("cls_A, cls_B, ancestor", [*general_ancestry()])
def test_closest_common_ancestor(cls_A, cls_B, ancestor):
    """Test closest common ancestor retrieval."""
    assert _closest_common_ancestor(cls_A, cls_B) is ancestor
    assert _closest_common_ancestor(cls_B, cls_A) is ancestor
    assert _closest_common_ancestor(cls_A(), cls_B) is ancestor
    assert _closest_common_ancestor(cls_B, cls_A()) is ancestor
    assert _closest_common_ancestor(cls_A, cls_B()) is ancestor
    assert _closest_common_ancestor(cls_B(), cls_A) is ancestor
    assert _closest_common_ancestor(cls_A(), cls_B()) is ancestor
    assert _closest_common_ancestor(cls_B(), cls_A()) is ancestor


@mark.parametrize("cls_A, cls_B, ancestor", [*strategy_ancestry()])
def test_shared_strategy_ancestor(cls_A, cls_B, ancestor):
    """Test strategy ancestor retrieval."""
    assert _shared_strategy_ancestor(cls_A, cls_B) is ancestor
    assert _shared_strategy_ancestor(cls_B, cls_A) is ancestor
    assert _shared_strategy_ancestor(cls_A(), cls_B) is ancestor
    assert _shared_strategy_ancestor(cls_B, cls_A()) is ancestor
    assert _shared_strategy_ancestor(cls_A, cls_B()) is ancestor
    assert _shared_strategy_ancestor(cls_B(), cls_A) is ancestor
    assert _shared_strategy_ancestor(cls_A(), cls_B()) is ancestor
    assert _shared_strategy_ancestor(cls_B(), cls_A()) is ancestor


################################################################################
## STRATEGY CLASSES
################################################################################
@mark.parametrize(
    "init, namespace",
    [
        (lambda self: None, ()),
        (lambda self, foo: None, ("foo",)),
        (lambda self, _bar: None, ("_bar",)),
        (lambda self, foo, _bar: None, ("foo", "_bar")),
        (lambda self, foo, bar: None, ("foo", "bar")),
        (lambda self, _foo, _bar: None, ("_foo", "_bar")),
    ],
)
class TestStrategy:
    """Test Strategy class."""

    @mark.parametrize("cls_name", ["cls", "Class", "Klass"])
    def test_init_subclass(self, cls_name, init, namespace):
        """Test subclassing logic."""
        settings = tuple(n for n in namespace if not n.startswith("_"))
        cls = type(cls_name, (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        assert hasattr(cls, "NAME")
        assert cls.NAME == obj.NAME == cls_name
        assert hasattr(cls, "INIT_NAMESPACE")
        assert isinstance(cls.INIT_NAMESPACE, tuple)
        assert cls.INIT_NAMESPACE is obj.INIT_NAMESPACE
        assert cls.INIT_NAMESPACE == tuple(namespace)
        assert hasattr(cls, "SETTINGS_NAMESPACE")
        assert isinstance(cls.SETTINGS_NAMESPACE, tuple)
        assert cls.SETTINGS_NAMESPACE is obj.SETTINGS_NAMESPACE
        assert cls.SETTINGS_NAMESPACE == tuple(settings)

    def test_new(self, init, namespace):
        """Test constructor logic."""
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        assert hasattr(obj, "_original_args")
        assert obj._original_args == kwargs

    def test_getattr(self, init, namespace):
        """Test getattr."""
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        for n, attr in enumerate(namespace):
            assert getattr(obj, attr) == n
        for attr in ("non_attr", "error"):
            with raises(AttributeError):
                getattr(obj, attr)

    def test_super_getattr(self, init, namespace):
        """Test extending getattr."""
        _super = type("super", (), {"__getattr__": lambda self, attr: attr})
        cls = type("cls", (_BaseStrategy, _super), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        for attr in ("non_attr", "error"):
            assert getattr(obj, attr) == attr
        for _, attr in enumerate(namespace):
            assert getattr(obj, attr) == attr

    def test_repr(self, init, namespace):
        """Test string representation."""
        settings = tuple(n for n in namespace if not n.startswith("_"))
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        cls_name = cls.__name__
        settings_str = ", ".join(f"{s}={getattr(obj, s)}" for s in settings)
        str_repr = f"{cls_name}({settings_str})" if settings_str else cls_name
        assert str(obj) == repr(obj) == str_repr

    def test_eq(self, init, namespace):
        """Test basic equality (same class, different settings).

        Note: checks reflexivity, symmetry, and transitivity.
        """
        settings = tuple(n for n in namespace if not n.startswith("_"))
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        copy = cls(**kwargs)
        assert obj == obj  # Reflexivity
        assert obj == copy == obj  # Symmetry
        assert obj == copy == cls(**kwargs) == obj  # Transitivity
        if settings:  # Different settings
            kwargs = {name: i + 1 for i, name in enumerate(namespace)}
            assert cls(**kwargs) != obj

    def test_settings(self, init, namespace):
        """Test settings property."""
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        settings = tuple(n for n in namespace if not n.startswith("_"))
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        assert obj.settings == {n: i for i, n in enumerate(namespace) if n in settings}
        for n in settings:
            setattr(obj, n, n)
            assert obj.settings.get(n) == getattr(obj, n)
        assert obj.settings == {n: n for n in settings}

    def test_init_args(self, init, namespace):
        """Test init_args property."""
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        assert obj.init_args == {n: i for i, n in enumerate(namespace)}
        for n in namespace:
            setattr(obj, n, n)
            assert obj.init_args.get(n) == getattr(obj, n)
        assert obj.init_args == {n: n for n in namespace}

    def test_original_args(self, init, namespace):
        """Test original_args property"""
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        assert obj.original_args == obj._original_args == kwargs
        assert obj.original_args is not obj._original_args

    def test_replicate(self, init, namespace):
        """Test replicate method."""
        cls = type("cls", (_BaseStrategy,), {"__init__": init})
        kwargs = {name: i for i, name in enumerate(namespace)}
        obj = cls(**kwargs)
        updates_list = (
            {},
            {n: n for n in namespace},
            {n: n for n, _ in zip(namespace, range(1))},
            {n: n for n, _ in zip(namespace, range(2))},
        )
        for updates in updates_list:
            replica = obj.replicate(**updates)
            assert type(replica) is type(obj)
            assert replica is not obj
            (init_args := obj.original_args).update(**updates)
            assert init_args == replica.original_args


class TestStrategyNonParametric:
    """Test Strategy class without parametrization."""

    def test_eq_facade(self):
        """Test equality for strategy facades.

        Note: reflexivity assumed from `TestStrategy.test_eq`, symmetry always checked,
        and transitivity checked where relevant (specified in comments).

        Class-legend: {
            N: {},
            _BaseStrategy: {
                clz: {},
                cls: {A: {}, B: {BB: {}}, C: {}, _C: {}}
            }
        }
        """
        cls = type("cls", (_BaseStrategy,), {"__init__": lambda self, a, b: None})
        clz = type("clz", (_BaseStrategy,), {"__init__": lambda self, a, b: None})
        A = type("A", (cls,), {"__init__": lambda self, a: None, "b": "b"})
        B = type("B", (cls,), {"__init__": lambda self, b: None, "a": "a"})
        BB = type("BB", (B,), {"__init__": lambda self, b: None, "a": "a"})
        C = type("C", (cls,), {"__init__": lambda self, c: None, "a": "a", "b": "b"})
        _C = type("_C", (cls,), {"__init__": lambda self, _c: None, "a": "a", "b": "b"})
        N = type("N", (), {"__init__": lambda self, a, b: None})
        assert cls("a", "b") == A("a") == cls("a", "b")  # Facade
        assert cls("a", "b") == B("b") == cls("a", "b")  # Facade
        assert A("a") == B("b") == A("a")  # Different facades (transitivity)
        assert cls("a", "b") != cls("b", "a") != cls("a", "b")  # Different settings
        assert cls("a", "a") != A("a") != cls("a", "a")  # Different settings
        assert cls("a", "a") != B("b") != cls("a", "a")  # Different settings
        assert A("b") != B("a") != A("b")  # Different facades different settings
        assert cls("a", "b") != clz("a", "b") != cls("a", "b")  # Different strategies same settings
        assert cls("a", "b") != N("a", "b") != cls("a", "b")  # Non-strategy
        assert cls("a", "b") != C("c") != cls("a", "b")  # Non-facade
        assert cls("a", "b") == _C("_c") == cls("a", "b")  # Excluded extra setting
        assert cls("a", "b") == BB("b") == cls("a", "b")  # Sub-subclass
        assert BB("b") == _C("_c") == BB("b")  # Furthest (transitivity)

    @mark.parametrize("cls, ancestor, expected", [*strategy_facades()])
    def test_is_facade(self, cls, ancestor, expected):
        """Test strategy facade detection."""
        obj = cls(**{k: k for k in _infer_init_namespace(cls)})
        ancestor_obj = ancestor(**{k: k for k in _infer_init_namespace(ancestor)})
        assert cls.is_facade(ancestor) is expected
        assert obj.is_facade(ancestor) is expected
        assert cls.is_facade(ancestor_obj) is expected
        assert obj.is_facade(ancestor_obj) is expected
        if expected:  # Inverted is False
            expected = cls is ancestor  # Only `True` if equal
            assert ancestor.is_facade(cls) is expected
            assert ancestor.is_facade(obj) is expected
            assert ancestor_obj.is_facade(cls) is expected
            assert ancestor_obj.is_facade(obj) is expected

    @mark.parametrize(
        "init, error",
        [
            (lambda self: None, False),
            (lambda self, /: None, False),
            (lambda self, foo: None, False),
            (lambda self, /, foo: None, False),
            (lambda self, foo, /: None, True),
            (lambda self, foo, /, bar: None, True),
            (lambda self, foo, bar, /: None, True),
            (lambda self, *args: None, True),
            (lambda self, foo, *args: None, True),
            (lambda self, foo, *args, bar: None, True),
            (lambda self, foo, *, bar: None, False),
            (lambda self, *, foo, bar: None, False),
            (lambda self, **kwargs: None, True),
            (lambda self, foo, **kwargs: None, True),
            (lambda self, *, foo, **kwargs: None, True),
            (lambda self, foo, *, bar, **kwargs: None, True),
            (lambda self, *args, **kwargs: None, True),
            (lambda self, foo, /, *args, **kwargs: None, True),
        ],
    )
    def test_validate_init(self, init, error):
        if error:
            with raises(TypeError):
                _ = type("cls", (_BaseStrategy,), {"__init__": init})
        else:
            _ = type("cls", (_BaseStrategy,), {"__init__": init})



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered fr om the originals.

from unittest.mock import Mock, call, patch

from pytest import fixture, mark, raises

from zne.utils.docstrings import (
    check_section_exists_in_docstring,
    get_insert_location_in_docstring,
    get_num_leading_spaces,
    insert_into_docstring,
    insert_into_docstring_lines,
    insert_new_content_into_docstring,
    insert_new_section_into_docstring,
    remove_all_leading_spaces_in_docstring,
    validate_section,
)

################################################################################
## CONSTANTS
################################################################################

SECTIONS = ["Args", "Returns", "Yields", "Raises"]

MOCK_TARGET_PATH = "zne.utils.docstrings"

docstring1 = """Some explanation.

    And more.

    Args:
        param1: The first parameter.
        param2: The second parameter,
            some more explanation.

    Returns:
        True if successful, False otherwise.
    """

docstring2 = """Some explanation.

        Returns:
            True if successful, False otherwise.

        Raises:
            Error: If something fails.
        """

docstring3 = """Some explanation.

Yields:
    The next number.
"""

DOCSTRINGS = [docstring1, docstring2, docstring3]
DOCSTRINGS_LINES = [docstring.split("\n") for docstring in DOCSTRINGS]

################################################################################
## MODULE FIXTURES
################################################################################


@fixture(scope="module")
def docstring():
    return DOCSTRINGS[0]


@fixture(scope="module")
def docstring_lines():
    return DOCSTRINGS_LINES[0]


@fixture(scope="module")
def new_content():
    return "param: example"


@fixture(scope="module")
def section():
    return "Args"


@fixture(scope="module")
def tab_size():  # TODO use '\t' instead
    return 4


@fixture(scope="module")
def patch_with_mock():
    def factory_method(method_name=None, **kwargs):
        method_name = method_name or ""
        return patch(MOCK_TARGET_PATH + method_name, **kwargs)

    return factory_method


@fixture(scope="module")
def patch_with_multiple_mocks():
    def factory_method(**kwargs):
        return patch.multiple(MOCK_TARGET_PATH, **kwargs)

    return factory_method


################################################################################
## TESTS
################################################################################


def test_insert_into_docstring(patch_with_multiple_mocks, docstring):
    new_content_list = [("Args", "param"), ("Raises", "Error")]
    mocks = {
        "remove_all_leading_spaces_in_docstring": Mock(return_value=["string1"]),
        "validate_section": Mock(),
        "insert_into_docstring_lines": Mock(return_value=["string2", "string3"]),
    }
    with patch_with_multiple_mocks(**mocks):
        new_docstring = insert_into_docstring(docstring, new_content_list)
    mocks["remove_all_leading_spaces_in_docstring"].assert_called_once_with(docstring.split("\n"))
    assert mocks["validate_section"].call_count == len(new_content_list)
    calls = [call("Args"), call("Raises")]
    mocks["validate_section"].assert_has_calls(calls, any_order=False)
    assert mocks["insert_into_docstring_lines"].call_count == len(new_content_list)
    calls = [call(["string1"], "param", "Args"), call(["string2", "string3"], "Error", "Raises")]
    mocks["insert_into_docstring_lines"].assert_has_calls(calls, any_order=False)
    assert new_docstring == "string2\nstring3"


def test_remove_all_leading_spaces_in_docstring(patch_with_mock, docstring_lines):
    docstring_lines_copy = docstring_lines.copy()
    with patch_with_mock(".get_num_leading_spaces", return_value=3) as mock:
        new_docstring_lines = remove_all_leading_spaces_in_docstring(docstring_lines_copy)
    mock.assert_called_once_with(docstring_lines_copy)
    assert new_docstring_lines[0] == docstring_lines[0]
    for i, line in enumerate(new_docstring_lines[1:]):
        assert line == docstring_lines[i + 1][3:]


@mark.parametrize(
    "docstring_lines, num_leading_spaces_expected",
    cases := tuple(
        zip(
            DOCSTRINGS_LINES,
            [4, 8, 0],
        )
    ),
    ids=[f"{type(dl).__name__}<{len(dl)}>-{nlse}" for dl, nlse in cases],
)
def test_get_num_leading_spaces(docstring_lines, num_leading_spaces_expected):
    assert get_num_leading_spaces(docstring_lines) == num_leading_spaces_expected


@mark.parametrize("docstring_lines", cases := [[""], ["One-liner"]], ids=[f"{c}" for c in cases])
def test_get_num_leading_spaces_value_error(docstring_lines):
    with raises(ValueError):
        assert get_num_leading_spaces(docstring_lines)


@mark.parametrize("section", cases := SECTIONS, ids=[f"{c}" for c in cases])
def test_validate_section(section):
    validate_section(section)


@mark.parametrize("section", cases := ["Argss", "returns", "Raises:"], ids=[f"{c}" for c in cases])
def test_validate_section_value_error(section):
    with raises(ValueError):
        assert validate_section(section)


def test_insert_into_docstring_lines_section_exists_false(
    patch_with_multiple_mocks, section, new_content
):
    mocks = {
        "check_section_exists_in_docstring": Mock(return_value=False),
        "get_insert_location_in_docstring": Mock(return_value=1),
        "insert_new_section_into_docstring": Mock(return_value=(["string2"], 2)),
        "insert_new_content_into_docstring": Mock(return_value=(["string3"], 3)),
    }
    with patch_with_multiple_mocks(**mocks):
        docstring_lines = insert_into_docstring_lines(["string1"], new_content, section)
    mocks["check_section_exists_in_docstring"].assert_called_once_with(["string1"], section)
    mocks["get_insert_location_in_docstring"].assert_called_once_with(["string1"], section)
    mocks["insert_new_section_into_docstring"].assert_called_once_with(["string1"], 1, section)
    mocks["insert_new_content_into_docstring"].assert_called_once_with(["string2"], 2, new_content)
    assert docstring_lines == ["string3"]


def test_insert_into_docstring_lines_section_exists_true(
    patch_with_multiple_mocks, section, new_content
):
    mocks = {
        "check_section_exists_in_docstring": Mock(return_value=True),
        "get_insert_location_in_docstring": Mock(return_value=1),
        "insert_new_section_into_docstring": Mock(return_value=(["string2"], 2)),
        "insert_new_content_into_docstring": Mock(return_value=(["string3"], 3)),
    }
    with patch_with_multiple_mocks(**mocks):
        docstring_lines = insert_into_docstring_lines(["string1"], new_content, section)
    mocks["check_section_exists_in_docstring"].assert_called_once_with(["string1"], section)
    mocks["get_insert_location_in_docstring"].assert_called_once_with(["string1"], section)
    mocks["insert_new_section_into_docstring"].assert_not_called()
    mocks["insert_new_content_into_docstring"].assert_called_once_with(["string1"], 1, new_content)
    assert docstring_lines == ["string3"]


@mark.parametrize(
    "docstring_lines, section, section_exists_expected",
    cases := tuple(
        zip(
            DOCSTRINGS_LINES,
            ["Args", "Yields", "Raises"],
            [True, False, False],
        )
    ),
    ids=[f"{type(dl).__name__}<{len(dl)}>-{s}-{see}" for dl, s, see in cases],
)
def test_check_section_exists_in_docstring(docstring_lines, section, section_exists_expected):
    assert check_section_exists_in_docstring(docstring_lines, section) == section_exists_expected


@mark.parametrize(
    "docstring_lines, section, insert_location_expected",
    cases := tuple(
        zip(
            DOCSTRINGS_LINES,
            ["Args", "Yields", "Raises"],
            [8, 4, 4],
        )
    ),
    ids=[f"{type(dl).__name__}<{len(dl)}>-{s}-{ile}" for dl, s, ile in cases],
)
def test_get_insert_location_in_docstring(docstring_lines, section, insert_location_expected):
    assert get_insert_location_in_docstring(docstring_lines, section) == insert_location_expected


@mark.parametrize(
    "docstring_lines, section, insert_location",
    cases := tuple(
        zip(
            DOCSTRINGS_LINES,
            ["Args", "Yields", "Raises"],
            [8, 4, 4],
        )
    ),
    ids=[f"{type(dl).__name__}<{len(dl)}>-{s}-{il}" for dl, s, il in cases],
)
def test_insert_new_section_into_docstring(docstring_lines, section, insert_location):
    new_docstring_lines, new_insert_location = insert_new_section_into_docstring(
        docstring_lines.copy(), insert_location, section
    )
    assert new_insert_location == insert_location + 2
    assert new_docstring_lines[:insert_location] == docstring_lines[:insert_location]
    assert new_docstring_lines[insert_location] == ""
    assert new_docstring_lines[insert_location + 1] == section + ":"
    assert new_docstring_lines[insert_location + 2 :] == docstring_lines[insert_location:]


@mark.parametrize(
    "docstring_lines, new_content, insert_location",
    cases := tuple(
        zip(
            DOCSTRINGS_LINES,
            ["Lorem ipsum"],
            [8, 4, 4],
        )
    ),
    ids=[f"{type(dl).__name__}<{len(dl)}>-{nc}-{il}" for dl, nc, il in cases],
)
def test_insert_new_content_into_docstring_single_line(
    docstring_lines, new_content, insert_location, tab_size
):
    new_docstring_lines, new_insert_location = insert_new_content_into_docstring(
        docstring_lines.copy(), insert_location, new_content
    )
    new_content_lines = new_content.split("\n")
    expected_new_insert_location = insert_location + len(new_content_lines)
    assert new_insert_location == expected_new_insert_location
    assert len(new_docstring_lines) == len(docstring_lines) + len(new_content_lines)
    assert new_docstring_lines[:insert_location] == docstring_lines[:insert_location]
    assert new_docstring_lines[insert_location] == " " * tab_size + new_content_lines[0]
    assert new_docstring_lines[expected_new_insert_location:] == docstring_lines[insert_location:]


@mark.parametrize(
    "docstring_lines, new_content, insert_location",
    cases := tuple(
        zip(
            DOCSTRINGS_LINES,
            ["Lorem\nipsum", "dolor\nsit\namet"],
            [8, 4, 4],
        )
    ),
    ids=[f"{type(dl).__name__}<{len(dl)}>-{nc}-{il}" for dl, nc, il in cases],
)
def test_insert_new_content_into_docstring_multiple_lines(
    docstring_lines, new_content, insert_location, tab_size
):
    new_docstring_lines, new_insert_location = insert_new_content_into_docstring(
        docstring_lines.copy(), insert_location, new_content
    )
    new_content_lines = new_content.split("\n")
    expected_new_insert_location = insert_location + len(new_content_lines)
    assert new_insert_location == expected_new_insert_location
    assert len(new_docstring_lines) == len(docstring_lines) + len(new_content_lines)
    assert new_docstring_lines[:insert_location] == docstring_lines[:insert_location]
    assert new_docstring_lines[insert_location] == " " * tab_size + new_content_lines[0]
    expected = [" " * tab_size * 2 + content for content in new_content_lines[1:]]
    assert new_docstring_lines[insert_location + 1 : expected_new_insert_location] == expected
    assert new_docstring_lines[expected_new_insert_location:] == docstring_lines[insert_location:]
